"""
########################################################################
The visualization.py module contains the Visualization class, which
provides tools for visualizing pyfao56 Model output with measured soil
water data.

The visualization.py module contains the following:
    Visualization - A class for visualizing pyfao56 Model output

11/09/2022 Initial Python functions developed by Josh Brekel, USDA-ARS
07/27/2023 Separated utility functions into their own classes, JB
08/28/2023 Major overhaul for pyfao56 release
10/31/2023 Added crop coefficient plot
########################################################################
"""

import matplotlib.pyplot as plt
from datetime import datetime as dt

class Visualization:
    """A class for visualizing pyfao56 Model output with measurements.

    Attributes
    ----------
    mdl : pyfao56 Model class
        Provides the simulated data
    sws : pyfao56 SoilWaterSeries class, optional
        Provides the measured soil water series data
        (default = None)
    todayidx : str
        Year and day of year of today ('yyyy-ddd')
    vdata : Dataframe
        Combined DataFrame of simulated (mdl.odata) and measured
        (sws.summarize()) root zone soil water data

    Methods
    -------
    plot_Dr(drmax=False, raw=False, events=False, obs=False, ks=False,
            dpro=False, title='', show=True, filepath=None)
        Create a plot of simulated soil water depletion
    plot_ET(refET=True, ETcm=True, ETc=True, ETa=True, events=False,
            title='', show=True, filepath=None)
        Create a plot of simulated evapotranspiration
    plot_Kc(Kcm=True, tKcb=True, Kcb=True, Ke=True, Kc=True, Ka=True,
            title='', show=True, filepath=None)
        Create a plot of simulated crop coefficient data
    """

    def __init__(self, mdl, sws=None, dayline=False):
        """Initialize the Visualization class attributes.

        Parameters
        ----------
        mdl : pyfao56 Model object
            Provides the simulated data
        sws : pyfao56 SoilWaterSeries object, optional
            Provides the measured soil water series data
            (default = None)
        dayline : boolean
            Adds a vertical line on the day the plot is created
            (default = False)
        """

        self.mdl = mdl
        self.sws = sws

        #Remove duplicated odata columns: Year, DOY, DOW, Date
        columns=['Year','DOY','DOW','Date']
        odatasub = mdl.odata.drop(columns=columns)

        #Add measured data if available
        if self.sws is not None:
            self.vdata = odatasub.merge(self.sws.summarize(),
                                        right_index=True,
                                        left_index=True,
                                        how='outer')
        else:
            self.vdata = odatasub

        #Set zero rain, irrigation, DP, and runoff to NaN
        NaN = float('NaN')
        self.vdata['Rain'] = self.vdata['Rain'].replace(0.0,NaN)
        self.vdata['Irrig'] = self.vdata['Irrig'].replace(0.0,NaN)
        self.vdata['DP'] = self.vdata['DP'].replace(0.0,NaN)
        self.vdata['Runoff'] = self.vdata['Runoff'].replace(0.0,NaN)

        #Determine today's index
        self.todayidx = ''
        if dayline:
            now = dt.today()
            self.todayidx = now.strftime('%Y-%j')

    def plot_Dr(self, drmax=False, raw=False, events=False, obs=False,
                ks=False,dpro=False,title='',show=True,filepath=None):
        """Plot soil water depletion (Dr) and related water data.

        Parameters
        ----------
        drmax : boolean, optional
            If True, include a line plot of simulated Drmax
            (default = False)
        raw : boolean, optional
            If True, include a line plot of simulated RAW
            (default = False)
        events : boolean, optional
            If True, include a scatter plot of irrigation and rain
            (default = False)
        obs : boolean, optional
            If True, include a scatter plot of measured Dr
            (default = False)
        ks : boolean, optional
            If True, include a plot of Ks and mKs at top
            (default = False)
        dpro : boolean, optional
            If True, include a scatter plot of simulated DP & runoff
            (default = False)
        title : str, optional
            Specify the title as the provided string
            (default = '')
        show : boolean, optional
            If True, the plot is displayed on the screen
            (default = True)
        filepath : str, optional
            Provide a filepath string to save the figure
            (default = None)
        """

        #Find a dpro_max divisible by four
        if self.vdata[['DP','Runoff']].isnull().all().all():
            dpro_max = 0.0
        else:
            dpro_max = round(self.vdata[['DP','Runoff']].max().max())
        i = int(max([int(dpro_max)+1.0, 4.0]))
        while True:
            if (not i%4): break
            else: i+=1
        dpro_max=float(i)
        #Check plotting conditions to determine axes
        if dpro and ks:
            htrat = {'height_ratios':[4, 32, 6]}
            fig, (ax2,ax,ax3) = plt.subplots(3, sharex='all',
                                             gridspec_kw=htrat)
            ax3.yaxis.set_major_locator(plt.MaxNLocator(4))
        elif dpro and not ks:
            htrat = {'height_ratios':[16, 3]}
            fig, (ax,ax3) = plt.subplots(2, sharex='all',
                                         gridspec_kw=htrat)
            ax3.yaxis.set_major_locator(plt.MaxNLocator(4))
        elif not dpro and ks:
            htrat = {'height_ratios': [1, 8]}
            fig, (ax2,ax) = plt.subplots(2, sharex='all',
                                         gridspec_kw=htrat)
        else: #not dpro and not ks
            fig, ax = plt.subplots()

        #Continue setting up the figure
        fig.set_size_inches(9.0,6.0)
        plt.subplots_adjust(left=0.06, right=0.99, bottom=0.07,
                            top=0.95, hspace=0.00, wspace=0.00)
        font = 8

        #Define x from 0 to n days to permit spanning years
        d = self.vdata
        x = range(len(d.index))
        xticks = []
        xlabels = []
        vline = float('NaN')
        for i, idx in enumerate(d.index):
            doy = int(idx[-3:])
            if not doy%5:  #if DOY is divisible by 5
                xticks.append(i)
                if not doy%2: #Label by tens
                    xlabels.append(idx[-3:])
                else:
                    xlabels.append('')
            if idx == self.todayidx:
                vline = i

        #Find maximum water state for scaling y axis
        maxwat = [d['Dr'].max()]
        if drmax:
            maxwat.append(d['Drmax'].max())
        if raw:
            maxwat.append(d['RAW'].max())
        if events:
            maxwat.append(d['Rain'].max())
            maxwat.append(d['Irrig'].max())
        if self.sws is not None:
            maxwat.append(d['mDr'].max())
            maxwat.append(d['mDrmax'].max())
        maxwat = round(max(maxwat))

        #Create the Ks plot
        if ks:
            ks_c = 'lightsalmon'
            ax2.plot(x, d['Ks'], color=ks_c, label='Simulated Ks')
            if obs:
                ax2.scatter(x, d['mKs'], color=ks_c, marker='s',
                            s=40,edgecolor='salmon',label='Measured Ks')
            ax2.set_xlim([xticks[0]-5, xticks[-1]+5])
            ax2.set_xticks(xticks)
            ax2.set_xticklabels(xlabels, fontsize=font)
            ax2.set_xlabel('Day of Year (DOY)', fontsize=font)
            ax2.set_ylim([-0.3, 1.2])
            ax2.set_yticks([0.0, 0.5, 1.0])
            ax2.set_yticklabels(['0.0','0.5','1.0'], fontsize=font)
            ax2.set_ylabel('Ks', fontsize=font)
            ax2.grid(ls=':')
            ax2.set_facecolor('whitesmoke')
            ax2.legend(fontsize=font,loc='upper right',frameon=False)
            if vline is not float('NaN'):
                ax2.axvline(x=vline, color='red', linestyle='--',
                            linewidth=0.5)

        #Create the DP plot
        if dpro:
            ax3.scatter(x, d['DP'], color='crimson', marker=10,
                       s=60, label='Deep Percolation (DP)')
            ax3.scatter(x, d['Runoff'], color='blue', marker="o",
                       s=60, label='Runoff (RO)')
            ax3.set_xlim([xticks[0]-5, xticks[-1]+5])
            ax3.set_xticks(xticks)
            ax3.set_xticklabels(xlabels,fontsize=font)
            ax3.set_xlabel('Day of Year (DOY)', fontsize=font)
            ax3.set_ylim([0.0, dpro_max])
            yticks = [round(i) for i in ax3.get_yticks()][1:]
            ax3.set_yticks(yticks)
            yticklabels = [str(i) for i in yticks]
            ax3.set_yticklabels(yticklabels, fontsize=font)
            ax3.set_ylabel('DP & RO (mm)', fontsize=font)
            ax3.invert_yaxis()
            ax3.grid(ls=':')
            ax3.set_facecolor('whitesmoke')
            ax3.legend(fontsize=font, loc='lower right', frameon=False)
            if vline is not float('NaN'):
                ax3.axvline(x=vline, color='red', linestyle='--',
                            linewidth=0.5)

        #Create the main plot
        ax.plot(x,d['Dr'],color='darkcyan',
                label='Simulated Root Zone Depletion')
        if drmax:
            ax.plot(x, d['Drmax'], color='darkturquoise',
                    label='Simulated Max Root Zone Depletion')
        if raw:
            ax.plot(x, d['RAW'], color='mediumorchid',
                    label='Readily Available Water (RAW)')
        if obs:
            ax.scatter(x, d['mDr'], color='darkcyan',
                       marker='s', s=40, edgecolor='darkslategray',
                       label='Measured Root Zone Depletion')
        if obs and drmax:
            ax.scatter(x, d['mDrmax'], color='darkturquoise',
                       marker='s', s=40, edgecolor='teal',
                       label='Measured Max Root Zone Depletion')
        if events:
            ax.scatter(x, d['Rain'], color='navy', marker='+',
                       s=35, linewidth=0.70, label='Rain')
            ax.scatter(x, d['Irrig'], color='navy', marker='x',
                       s=35, linewidth=0.70, label='Irrigation')
        ax.set_xlim([xticks[0]-5., xticks[-1]+5.])
        ax.set_xticks(xticks)
        ax.set_xticklabels(xlabels,fontsize=font)
        ax.set_xlabel('Day of Year (DOY)', fontsize=font)
        ax.set_ylim([0.0, maxwat+5.])
        ax.set_yticks(range(0, maxwat+5, 5))
        ax.set_yticklabels(range(0, maxwat+5, 5),fontsize=font)
        ax.set_ylabel('Depth (mm)', fontsize=font)
        ax.grid(ls=":")
        ax.set_facecolor('whitesmoke')
        ax.legend(fontsize=font, loc='upper left', frameon=False)
        if vline is not float('NaN'):
            ax.axvline(x=vline, color='red', linestyle='--',
                       linewidth=0.5)
        plt.suptitle(title, fontsize=10)

        #Save and show the plot if requested
        if filepath is not None:
            plt.savefig(filepath)
        if show:
            plt.show()
        else:
            plt.close(fig)

    def plot_ET(self, refET=True, ETcm=True, ETc=True, ETa=True,
                events=False, title='', show=True, filepath=None):
        """Plot evapotranspiration data versus time.

        Parameters
        ----------
        refET : boolean, optional
            If True, include a lineplot for reference ET
            (default = True)
        ETcm : boolean, optional
            If True, include a lineplot for ETcm (single method)
            (default = True)
        ETc : boolean, optional
            If True, include a lineplot for ETc (dual method)
            (default = True)
        ETa : boolean, optional
            If True, include a lineplot for actual ET
            (default = True)
        events : boolean, optional
            If True, include a scatter plot of irrigation and rain
            (default = False)
        title : str, optional
            Specify the title as the provided string
            (default = '')
        show : boolean, optional
            If True, the plot is displayed on the screen
            (default = True)
        filepath : str, optional
            Provide a filepath string to save the figure
            (default = None)
        """

        #Create the figure
        fig, ax = plt.subplots()
        fig.set_size_inches(9.0,6.0)
        plt.subplots_adjust(left=0.06, right=0.99, bottom=0.07,
                            top=0.95, hspace=0.00, wspace=0.00)
        font = 8

        #Define x from 0 to n days to permit spanning years
        d = self.vdata
        x = range(len(d.index))
        xticks = []
        xlabels = []
        vline = float('NaN')
        for i, idx in enumerate(d.index):
            doy = int(idx[-3:])
            if not doy%5: #if DOY is divisible by 5
                xticks.append(i)
                if not doy%2: #Label by tens
                    xlabels.append(idx[-3:])
                else:
                    xlabels.append('')
            if idx == self.todayidx:
                vline = i

        #Find maximum water state for scaling y axis
        maxwat = [d['ETref'].max(),d['ETcm'].max(),d['ETc'].max(),
                  d['ETa'].max()]
        if events:
            maxwat.append(d['Rain'].max())
            maxwat.append(d['Irrig'].max())
        maxwat = round(max(maxwat))

        #Create ET plot
        if refET:
            ax.plot(x, d['ETref'], color='darkred',
                    label='Reference ET (ETref)')
        if ETcm:
            ax.plot(x, d['ETcm'], color='blueviolet',
                    label='Non-stressed ET (ETcm, single)')
        if ETc:
            ax.plot(x, d['ETc'], color='blue',
                    label='Non-stressed ET (ETc, dual)')
        if ETa:
            ax.plot(x, d['ETa'], linestyle='-.', color='navy',
                    label='Actual ET (ETa)')
        if events:
            ax.scatter(x, d['Rain'], color='navy', marker='+',
                       s=35, linewidth=0.70, label='Rain')
            ax.scatter(x, d['Irrig'], color='navy', marker='x',
                       s=35, linewidth=0.70, label='Irrigation')

        ax.set_xlim([xticks[0]-5., xticks[-1]+5.])
        ax.set_xticks(xticks)
        ax.set_xticklabels(xlabels,fontsize=font)
        ax.set_xlabel('Day of Year (DOY)', fontsize=font)
        ax.set_ylim([0.0, maxwat+2.])
        yticks = [round(i) for i in ax.get_yticks()]
        ax.set_yticks(yticks)
        yticklabels = [str(i) for i in yticks]
        ax.set_yticklabels(yticklabels, fontsize=font)
        ax.set_ylabel('Depth (mm)', fontsize=font)
        ax.grid(ls=":")
        ax.set_facecolor('whitesmoke')
        ax.legend(fontsize=font, loc='upper left', frameon=False)
        if vline is not float('NaN'):
            ax.axvline(x=vline, color='red', linestyle='--',
                       linewidth=0.5)
        plt.suptitle(title, fontsize=10)

        #Save and show the plot if requested
        if filepath is not None:
            plt.savefig(filepath)
        if show:
            plt.show()
        else:
            plt.close(fig)

    def plot_Kc(self, Kcm=True, tKcb=True, Kcb=True, Ke=True, Kc=True,
                Ka=True, title='', show=True, filepath=None):
        """Plot crop coeffient data versus time.

        Parameters
        ----------
        Kcm : boolean, optional
            If True, include a plot of Kcm (single crop coefficient)
            (default = True)
        tKcb : boolean, optional
            If True, include a plot of tabular (trapezoidal) Kcb
            (default = True)
        Kcb : boolean, optional
            If True, include a plot of Kcb
            (default = True)
        Ke : boolean, optional
            If True, include a plot of Ke
            (default = True)
        Kc : boolean, optional
            If True, include a plot of Kc (dual crop coefficient)
            (default = True)
        Ka : boolean, optional
            If True, include a plot of Ka
            (default = True)
        title : str, optional
            Specify the title as the provided string
            (default = '')
        show : boolean, optional
            If True, the plot is displayed on the screen
            (default = True)
        filepath : str, optional
            Provide a filepath string to save the figure
            (default = None)
        """

        #Create the figure
        fig, ax = plt.subplots()
        fig.set_size_inches(9.0,6.0)
        plt.subplots_adjust(left=0.06, right=0.99, bottom=0.07,
                            top=0.95, hspace=0.00, wspace=0.00)
        font = 8

        # Define x from 0 to n days to permit spanning years
        d = self.vdata
        x = range(len(d.index))
        xticks = []
        xlabels = []
        vline = float('NaN')
        for i, idx in enumerate(d.index):
            doy = int(idx[-3:])
            if not doy%5:  #if DOY is divisible by 5
                xticks.append(i)
                if not doy%2: #Label by tens
                    xlabels.append(idx[-3:])
                else:
                    xlabels.append('')
            if idx == self.todayidx:
                vline = i

        #Create crop coefficient plot
        maxkc = 1.2
        if Kcm:
            kcm_c = 'blueviolet'
            ax.plot(x, d['Kcm'], color=kcm_c, label='Kcm')
            maxkc = round(max([maxkc,d['Kcm'].max()])+0.05,1)
        if tKcb:
            tkcb_c = 'mediumseagreen'
            ax.plot(x, d['tKcb'], linestyle='--', color=tkcb_c,
                    label='Tabular Kcb')
            maxkc = round(max([maxkc,d['tKcb'].max()])+0.05,1)
        if Kcb:
            kcb_c = 'seagreen'
            ax.plot(x, d['Kcb'], color=kcb_c, label='Kcb')
            maxkc = round(max([maxkc,d['Kcb'].max()])+0.05,1)
        if Ke:
            ke_c = 'lightskyblue'
            ax.plot(x, d['Ke'], color=ke_c, label='Ke')
            maxkc = round(max([maxkc,d['Ke'].max()])+0.05,1)
        if Kc:
            kc_c = 'blue'
            ax.plot(x, d['Kc'], color=kc_c, label='Kc')
            maxkc = round(max([maxkc,d['Kc'].max()])+0.05,1)
        if Ka:
            ka_c = 'dimgrey'
            ax.plot(x, d['Ka'], color=ka_c, label='Ka')
            maxkc = round(max([maxkc,d['Ka'].max()])+0.05,1)

        yticks=[]
        ylabels=[]
        for i in range(int(maxkc*10.)+1):
            yticks.append(float(i)/10.)
            ylabels.append(str(float(i)/10.))

        ax.set_xlim([xticks[0]-5., xticks[-1]+5.])
        ax.set_xticks(xticks)
        ax.set_xticklabels(xlabels, fontsize=font)
        ax.set_xlabel('Day of Year (DOY)', fontsize=font)
        ax.set_ylim([0.0, maxkc])
        ax.set_yticks(yticks)
        ax.set_yticklabels(ylabels, fontsize=font)
        ax.set_ylabel('Crop Coefficients', fontsize=font)
        ax.grid(ls=':')
        ax.set_facecolor('whitesmoke')
        ax.legend(fontsize=font,loc='upper right',frameon=False)
        if vline is not float('NaN'):
            ax.axvline(x=vline, color='red', linestyle='--',
                       linewidth=0.5)
        plt.suptitle(title, fontsize=10)

        #Save and show the plot if requested
        if filepath is not None:
            plt.savefig(filepath)
        if show:
            plt.show()
        else:
            plt.close(fig)



"""
########################################################################
The statistics.py module contains functions for computing fit statistics
between measured and simulated data, including
 1. Bias ('bias')
 2. Relative bias ('rbias')
 3. Percent bias ('pbias')
 4. Maximum error ('maxerr')
 5. Mean error ('meanerr')
 6. Mean absolute error ('mae')
 7. Sum of squared error ('sse')
 8. Pearson's correlation coefficient ('r')
 9. Coefficient of determination ('r2')
10. Root mean squared error ('rmse')
11. Relative root mean squared error ('rrmse')
12. Percent root mean squared error ('prmse')
13. Coefficient of residual mass ('crm')
14. Nash & Sutcliffe (1970) model efficiency ('nse')
15. Willmott (1981) index of agreement ('d')
16. Kling-Gupta (2011) efficiency ('kge')

11/01/2023 Initial Python script
08/05/2024 Added Kling-Gupta efficiency
########################################################################
"""

import numpy as np
import datetime

class Statistics:
    """A class for computing goodness-of-fit statistics

    Attributes
    ----------
    simulated : numpy array
        A 1d array of simulated data
    measured : numpy array
        A 1d array of measured data
    stats - dict
        Container for goodness-of-fit statistics
        keys - ['bias','rbias','pbias','maxerr','meanerr','mae','sse',
                'r','r2','rmse','rrmse','prmse','crm','nse','d','kge']
        value - Computed value of the statistic

    Methods
    -------
    savefile(filepath='pyfao56.fit')
        Save goodness-of-fit statistics to a file
    """

    def __init__(self, simulated, measured, comment=''):
        """Initialize the Statistics class attributes.

        Parameters
        ----------
        simulated : list
            A list of simulated data
        measured : list
            A list of measured data
        """

        self.comment = 'Comments: ' + comment.strip()
        self.simulated = np.array(simulated).flatten()
        self.measured = np.array(measured).flatten()
        s = self.simulated
        m = self.measured
        self.stats = {}
        self.stats.update({'bias'   :self._bias(s,m)})
        self.stats.update({'rbias'  :self._rbias(s,m)})
        self.stats.update({'pbias'  :self._pbias(s,m)})
        self.stats.update({'maxerr' :self._maxerr(s,m)})
        self.stats.update({'meanerr':self._meanerr(s,m)})
        self.stats.update({'mae'    :self._mae(s,m)})
        self.stats.update({'sse'    :self._sse(s,m)})
        self.stats.update({'r'      :self._r(s,m)})
        self.stats.update({'r2'     :self._r2(s,m)})
        self.stats.update({'rmse'   :self._rmse(s,m)})
        self.stats.update({'rrmse'  :self._rrmse(s,m)})
        self.stats.update({'prmse'  :self._prmse(s,m)})
        self.stats.update({'crm'    :self._crm(s,m)})
        self.stats.update({'nse'    :self._nse(s,m)})
        self.stats.update({'d'      :self._d(s,m)})
        self.stats.update({'kge'    :self._kge(s,m)})

    def __str__(self):
        """Represent the Statistics class variables as a string."""

        self.tmstmp = datetime.datetime.now()
        timestamp = self.tmstmp.strftime('%m/%d/%Y %H:%M:%S')
        ast='*'*72
        s = ('{:s}\n'
             'pyfao56: FAO-56 Evapotranspiration in Python\n'
             'Goodness-of-fit Statistics\n'
             'Timestamp: {:s}\n'
             '{:s}\n'
             '{:s}\n'
             '{:s}\n'
             ).format(ast,
                      timestamp,
                      ast,
                      self.comment,
                      ast)
        keys = ['bias','rbias','pbias','maxerr','meanerr','mae','sse',
                'r','r2','rmse','rrmse','prmse','crm','nse','d','kge']
        for key in keys:
            s += '{:s} : {:f}\n'.format(key, self.stats[key])
        return s

    def savefile(self,filepath='pyfao56.fit'):
        """Save goodness-of-fit statistics to a file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.fit')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'w')
        except FileNotFoundError:
            print('The filepath for output data is not found.')
        else:
            f.write(self.__str__())
            f.close()

    def _bias(self,s,m):
        """Compute the bias."""
        return np.sum(s-m)

    def _rbias(self,s,m):
        """Compute the relative bias."""
        return np.sum(s-m)/np.mean(m)

    def _pbias(self,s,m):
        """Compute the percent bias."""
        return np.sum(s-m)/np.mean(m)*100.

    def _maxerr(self,s,m):
        """Compute maximum error."""
        return np.max(np.absolute(s-m))

    def _meanerr(self,s,m):
        """Compute the mean error."""
        return np.mean(s-m)

    def _mae(self,s,m):
        """Compute the mean absolute error."""
        return np.mean(np.absolute(s-m))

    def _sse(self,s,m):
        """Compute the sum of squared error."""
        return np.sum(np.square(s-m))

    def _r(self,s,m):
        """Compute the Pearson correlation coefficient (r)."""
        a = np.sum((s-np.mean(s))*(m-np.mean(m)))
        b = np.sum(np.square(s-np.mean(s)))
        c = np.sum(np.square(m-np.mean(m)))
        return a/np.sqrt(b*c)
        #return np.corrcoef(s,m)[0][1]

    def _r2(self,s,m):
        """Compute the coefficient of determination (r^2)."""
        a = np.sum((s-np.mean(s))*(m-np.mean(m)))
        b = np.sum(np.square(s-np.mean(s)))
        c = np.sum(np.square(m-np.mean(m)))
        return (a/np.sqrt(b*c))**2.0

    def _rmse(self,s,m):
        """Compute the root mean squared error."""
        return np.sqrt(np.mean(np.square(s-m)))

    def _rrmse(self,s,m):
        """Compute the relative root mean squared error."""
        return np.sqrt(np.mean(np.square(s-m)))/np.mean(m)

    def _prmse(self,s,m):
        """Compute the percent root mean squared error."""
        return np.sqrt(np.mean(np.square(s-m)))/np.mean(m)*100.

    def _crm(self,s,m):
        """Compute the coefficient of residual mass."""
        return (np.sum(s)-np.sum(m))/np.sum(m)

    def _nse(self,s,m):
        """Compute the Nash & Sutcliffe (1970) model efficiency."""
        a = np.sum(np.square(s-m))
        b = np.sum(np.square(m-np.mean(m)))
        return 1.0-a/b

    def _d(self,s,m):
        """Compute the Willmott (1981) index of agreement (d)."""
        a = np.sum(np.square(s-m))
        b = np.absolute(s-np.mean(m))
        c = np.absolute(m-np.mean(m))
        d = np.sum(np.square(b+c))
        return 1.0-a/d

    def _kge(self,s,m):
        """Compute the Kling-Gupta (2011) efficiency (kge)."""
        r = self._r(s,m)
        beta = np.mean(s)/np.mean(m)
        alpha = np.std(s)/np.std(m)
        a = (r-1.0)*(r-1.0)
        b = (alpha-1.0)*(alpha-1.0)
        c = (beta-1.0)*(beta-1.0)
        return 1.0-np.sqrt(a+b+c)



"""
########################################################################
The pyfao56 Python package facilitates FAO-56 computations of daily soil
water balance using the dual crop coefficient method to estimate crop
evapotranspiration (ET).

The FAO-56 method is described in the following documentation:
Allen, R. G., Pereira, L. S., Raes, D., Smith, M., 1998.  FAO Irrigation
and Drainage Paper No. 56. Crop Evapotranspiration: Guidelines for
Computing Crop Water Requirements. Food and Agriculture Organization of
the United Nations, Rome Italy.

http://www.fao.org/3/x0490e/x0490e00.htm


Reference ET is computed using the ASCE Standardized Reference ET
Equation, which is described in the following documentation:
ASCE Task Committee on Standardization of Reference Evapotranspiration
(Walter, I. A., Allen, R. G., Elliott, R., Itenfisu, D., Brown, P.,
Jensen, M. E.,Mecham, B., Howell, T. A., Snyder, R., Eching, S.,
Spofford, T., Hattendorf, M., Martin, D., Cuenca, R. H., Wright, J. L.)
, 2005. The ASCE Standardized Reference Evapotranspiration Equation.
American Society of Civil Engineers, Reston, VA.

https://ascelibrary.org/doi/book/10.1061/9780784408056


The pyfao56 package contains the following modules:
    autoirrigate.py
        I/O tools for defining conditions for automatic irrigation
    irrigation.py
        I/O tools to define irrigation management schedules explicitly
    model.py
        Equations for daily soil water balance computations
    parameters.py
        I/O tools for required input parameters
    soil_profile.py
        I/O tools for defining soil properties stratified by soil layer
    refet.py
        Equations for computing ASCE Standardized Reference ET
    update.py
        I/O tools and methods for state variable updating
    weather.py
        I/O tools for required weather information

01/07/2016 Initial Python functions developed by Kelly Thorp
11/04/2021 Scripts updated for inclusion in the pyfao56 Python package
########################################################################
"""

from .autoirrigate import AutoIrrigate
from .irrigation import Irrigation
from .model import Model
from .parameters import Parameters
from .soil_profile import SoilProfile
from .update import Update
from .weather import Weather
from .__version__ import __version__



"""
########################################################################
The pyfao56 Python package facilitates FAO-56 computations of daily soil
water balance using the dual crop coefficient method to estimate crop
evapotranspiration (ET).

The FAO-56 method is described in the following documentation:
Allen, R. G., Pereira, L. S., Raes, D., Smith, M., 1998.  FAO Irrigation
and Drainage Paper No. 56. Crop Evapotranspiration: Guidelines for
Computing Crop Water Requirements. Food and Agriculture Organization of
the United Nations, Rome Italy.

http://www.fao.org/3/x0490e/x0490e00.htm


Reference ET is computed using the ASCE Standardized Reference ET
Equation, which is described in the following documentation:
ASCE Task Committee on Standardization of Reference Evapotranspiration
(Walter, I. A., Allen, R. G., Elliott, R., Itenfisu, D., Brown, P.,
Jensen, M. E.,Mecham, B., Howell, T. A., Snyder, R., Eching, S.,
Spofford, T., Hattendorf, M., Martin, D., Cuenca, R. H., Wright, J. L.)
, 2005. The ASCE Standardized Reference Evapotranspiration Equation.
American Society of Civil Engineers, Reston, VA.

https://ascelibrary.org/doi/book/10.1061/9780784408056

01/31/2025 Setup up __version__ methodology
########################################################################
"""

__version__ = "1.4.1"



"""
########################################################################
The model.py module contains the Model class, which defines the
equations for daily soil water balance calculations based on the FAO-56
dual crop coefficient method for evapotranspiration (ET) estimation.

The FAO-56 method is described in the following documentation:
Allen, R. G., Pereira, L. S., Raes, D., Smith, M., 1998.  FAO Irrigation
and Drainage Paper No. 56. Crop Evapotranspiration: Guidelines for
Computing Crop Water Requirements. Food and Agriculture Organization of
the United Nations, Rome, Italy.

http://www.fao.org/3/x0490e/x0490e00.htm

Further details on the FAO56 methodology (as well as the runoff method)
can be found in the following documentation:
ASCE Task Committee on Revision of Manual 70, 2016. Evaporation,
Evapotranspiration, and Irrigation Water Requirements, 2nd edition. ASCE
Manuals and Reports on Engineering Practice No. 70. Jensen, M. E. and
Allen, R. G. (eds.). American Society of Civil Engineers, Reston,
Virginia.

Current updates to ET terminology, which were incorporated herein, can
be found in the following documentation:
DeJonge, K. C., Allen, R. G., Kilic, A., Thorp, K. R., Kukal, M. S.,
Marek, G. W., Altenhofen, J., Amatya, D. M., Blankenau, P. A.,
Datta, S., Grabow, G. L., Hashem, A. A., Kisekka, I., Kjaersgaard, J.,
Marek, T. H., Peters, T. R., Porter, D. O., Reba, M. L., Rudnick, D. R.,
Senay, G. B., Sharma, V., Sridhar, V., Sun, G., Taghvaeian, S.,
Trezza, R., Trout, T. J., 2025. Evapotranspiration terminology and
definitions. Journal of Irrigation and Drainage Engineering. In review.


The model.py module contains the following:
    Model - A class for managing FAO-56 soil water balance computations.

01/07/2016 Initial Python functions developed by Kelly Thorp
11/04/2021 Finalized updates for inclusion in the pyfao56 Python package
10/27/2022 Incorporated Fort Collins ARS stratified soil layers approach
11/30/2022 Incorporated Fort Collins ARS water balance approach
08/17/2023 Improved logic for case of missing rhmin data
10/31/2023 Added AquaCrop Ks option
11/01/2023 Added reports of the cumulative seasonal water balance
12/12/2023 Added the runoff functionality by Dinesh Gulati
02/15/2024 Added functionality for automatic irrigation scheduling
09/30/2024 Added single crop coefficient calculations
01/29/2025 Added Kcb weather-based adjustment method (FAO-56 Eq. 70)
02/04/2025 Updated ET terminology based on DeJonge et al. (2025)
07/09/2025 Added Ksend computation for end of timestep
########################################################################
"""

import pandas as pd
import datetime
import math
from .__version__ import __version__

class Model:
    """A class for managing FAO-56 soil water balance computations.

    Manages computations based on FAO-56 methods for evapotranspiration
    and soil water balance calculations (Allen et al., 1998).

    Attributes
    ----------
    startDate : datetime
        Simulation start date in datetime format
    endDate : datetime
        Simulation end date in datetime format
    par : pyfao56 Parameters class
        Provides the parameter data for simulations
    wth : pyfao56 Weather class
        Provides the weather data for simulations
    irr : pyfao56 Irrigation class, optional
        Provides the irrigation data for simulations
        (default = None)
    sol : pyfao56 SoilProfile class, optional
        Provides data for modeling with stratified soil layers
        (default = None)
    autoirr : pyfao56 AutoIrrigate class, optional
        Provides data for automatic irrigation scheduling
        (default = None)
    upd : pyfao56 Update class, optional
        Provides data and methods for state variable updating
        (default = None)
    roff : boolean, optional
        If True, computes surface runoff following ASCE (2016)
        (default = False)
    cons_p : boolean, optional
        If False, p follows FAO-56; if True, p is constant (=pbase)
        (default = False)
    aq_Ks : boolean, optional
        If False, Ks follows FAO-56; if True, Ks via AquaCrop equation
        (default = False)
    K_adj : boolean, optional
        If True, Kcmmid, Kcmend, Kcbmid, and Kcbend are adjusted for
        weather following FAO-56 Eqs. 62, 65, and 70
        (default = False)
    comment : str, optional
        User-defined file descriptions or metadata (default = '')
    tmstmp : datetime
        Time stamp for the class
    ModelState : class
        Contains parameters and model states for a single timestep
    cnames : list
        Column names for odata
    odata : DataFrame
        Model output data as float
        index - Year and day of year as string ('yyyy-ddd')
        columns - ['Year','DOY','DOW','Date','ETref','Kcm','ETcm',
                   'tKcb','Kcb','ETcb','h','Kcmax','ETmax','fc','fw',
                   'few','De','Kr','Ke','E','DPe','Kc','ETc','TAW',
                   'TAWrmax','TAWb','Zr','p','RAW','Ks','Ka','ETa','T',
                   'DP','Dinc','Dr','fDr','Drmax','fDrmax','Db','fDb',
                   'Ksend','Irrig','IrrLoss','Rain','Runoff','Year',
                   'DOY','DOW','Date']
            Year    - 4-digit year (yyyy)
            DOY     - Day of year (ddd)
            DOW     - Day of week
            Date    - Month/Day/Year (mm/dd/yy)
            ETref   - Daily reference evapotranspiration (mm)
            Kcm     - Single (mean) crop coefficient, FAO-56 Eq. 56
            ETcm    - Single (mean) non-stressed ET (mm), FAO-56 Eq. 56
            tKcb    - Basal crop coefficient, trapezoidal from FAO-56
            Kcb     - Basal crop coefficient, considering updates
            ETcb    - Basal evapotranspiration (mm)
            h       - Plant height (m)
            Kcmax   - Maximum crop coefficient, FAO-56 Eq. 72
            ETmax   - Maximum evapotranspiration (mm)
            fc      - Canopy cover fraction, FAO-56 Eq. 76
            fw      - Fraction soil surface wetted, FAO-56 Table 20
            few     - Exposed & wetted soil fraction, FAO-56 Eq. 75
            De      - Cumulative depth of evaporation, FAO-56 Eqs. 77&78
            Kr      - Evaporation reduction coefficient, FAO-56 Eq. 74
            Ke      - Evaporation coefficient, FAO-56 Eq. 71
            E       - Soil water evaporation (mm), FAO-56 Eq. 69
            DPe     - Percolation under exposed soil (mm), FAO-56 Eq. 79
            Kc      - Dual non-stressed crop coefficient, FAO-56 Eq. 69
            ETc     - Dual non-stressed ET (mm), FAO-56 Eq. 69
            TAW     - Total available water (mm), FAO-56 Eq. 82
            TAWrmax - Total available water for max root depth (mm)
            TAWb    - Total available water in bottom layer (mm)
            Zr      - Root depth (m), FAO-56 page 279
            p       - Fraction depleted TAW, FAO-56 p162 and Table 22
            RAW     - Readily available water (mm), FAO-56 Equation 83
            Ks      - Stress coefficient, timestep start, FAO-56 Eq. 84
            Ka      - Actual crop coefficient, FA0-56 Eq. 80
            ETa     - Actual evapotranspiration (mm), FAO-56 Eq. 80
            T       - Actual plant transpiration (mm)
            DP      - Deep percolation (mm), FAO-56 Eq. 88
            Dinc    - Depletion increment due to root growth (mm)
            Dr      - Soil water depletion (mm), FAO-56 Eqs. 85 & 86
            fDr     - Fractional root zone soil water depletion (mm/mm)
            Drmax   - Soil water depletion for max root depth (mm)
            fDrmax  - Fractional depletion for max root depth (mm/mm)
            Db      - Soil water depletion in the bottom layer (mm)
            fDb     - Fractional depletion in the bottom layer (mm/mm)
            Ksend   - Stress coefficient, timestep end, FAO-56, Eq. 84
            Irrig   - Depth of applied irrigation (mm)
            IrrLoss - Depth of irrigation loss due to inefficiency (mm)
            Rain    - Depth of precipitation (mm)
            Runoff  - Surface runoff (mm)
            Year    - 4-digit year (yyyy)
            DOY     - Day of year (ddd)
            DOW     - Day of week
            Date    - Month/Day/Year (mm/dd/yy)
    swbdata : dict
        Container for cumulative seasonal water balance data
        keys - ['ETref','ETcm','ETcb','ETmax','ETc','ETa','E','T','DP',
                'Irrig','IrrLoss','Rain','Runoff','Dr_ini','Dr_end',
                'Drmax_ini','Drmax_end']
        value - Cumulative water balance data in mm

    Methods
    -------
    savefile(filepath='pyfao56.out')
        Save pyfao56 output data to a file
    savesums(filepath='pyfao56.sum')
        Save seasonal water balance data to a file
    run()
        Conduct the FAO-56 calculations from startDate to endDate
    """

    def __init__(self, start, end, par, wth, irr=None, autoirr=None,
                 sol=None, upd=None, roff=False, cons_p=False,
                 aq_Ks=False, K_adj=False, comment=''):
        """Initialize the Model class attributes.

        Parameters
        ----------
        start : str
            Simulation start year and doy ('yyyy-ddd')
        end : str
            Simulation end year and doy ('yyyy-ddd')
        par : pyfao56 Parameters object
            Provides the parameter data for simulations
        wth : pyfao56 Weather object
            Provides the weather data for simulations
        irr : pyfao56 Irrigation object, optional
            Provides the irrigation data for simulations
            (default = None)
        autoirr : pyfao56 AutoIrrigate object, optional
            Provides data for automatic irrigation scheduling
            (default = None)
        sol : pyfao56 SoilProfile object, optional
            Provides data for modeling with stratified soil layers
            (default = None)
        upd : pyfao56 Update object, optional
            Provides data and methods for state variable updating
            (default = None)
        roff : boolean, optional
            If True, computes surface runoff following ASCE (2016)
            (default = False)
        cons_p : boolean, optional
            If False, p follows FAO-56; if True, p is constant (=pbase)
            (default = False)
        aq_Ks : boolean, optional
            If False, Ks follows FAO-56; if True, Ks via AquaCrop Eqn
            (default = False)
        K_adj : boolean, optional
            If True, Kcmmid, Kcmend, Kcbmid, and Kcbend are adjusted for
            weather following FAO-56 Eqs. 62, 65, and 70
            (default = False)
        comment : str, optional
            User-defined file descriptions or metadata (default = '')
        """

        self.startDate = datetime.datetime.strptime(start, '%Y-%j')
        self.endDate   = datetime.datetime.strptime(end, '%Y-%j')
        self.par = par
        self.wth = wth
        self.irr = irr
        self.autoirr = autoirr
        self.sol = sol
        self.upd = upd
        self.roff = roff
        self.cons_p = cons_p
        self.aq_Ks = aq_Ks
        self.K_adj = K_adj
        self.comment = 'Comments: ' + comment.strip()
        self.tmstmp = datetime.datetime.now()
        self.cnames = ['Year','DOY','DOW','Date','ETref','Kcm','ETcm',
                       'tKcb','Kcb','ETcb','h','Kcmax','ETmax','fc','fw',
                       'few','De','Kr','Ke','E','DPe','Kc','ETc','TAW',
                       'TAWrmax','TAWb','Zr','p','RAW','Ks','Ka','ETa',
                       'T','DP','Dinc','Dr','fDr','Drmax','fDrmax','Db',
                       'fDb','Ksend','Irrig','IrrLoss','Rain','Runoff',
                       'Year','DOY','DOW','Date']
        self.odata = pd.DataFrame(columns=self.cnames)

    def __str__(self):
        """Represent the Model class variables as a string."""

        self.tmstmp = datetime.datetime.now()
        timestamp = self.tmstmp.strftime('%m/%d/%Y %H:%M:%S')
        sdate = self.startDate.strftime('%m/%d/%Y')
        edate = self.endDate.strftime('%m/%d/%Y')
        if self.sol is None:
            solmthd = 'D - Default FAO-56 homogenous soil bucket ' \
                      'approach'
        else:
            solmthd = 'L - Fort Collins ARS stratified soil layers ' \
                      'approach'
        fmts = {'Year':'{:4s}'.format,'DOY':'{:3s}'.format,
                'DOW':'{:3s}'.format,'Date':'{:8s}'.format,
                'ETref':'{:6.3f}'.format,'Kcm':'{:5.3f}'.format,
                'ETcm':'{:6.3f}'.format,'tKcb':'{:5.3f}'.format,
                'Kcb':'{:5.3f}'.format,'ETcb':'{:6.3f}'.format,
                'h':'{:5.3f}'.format,'Kcmax':'{:5.3f}'.format,
                'ETmax':'{:6.3f}'.format,'fc':'{:5.3f}'.format,
                'fw':'{:5.3f}'.format,'few':'{:5.3f}'.format,
                'De':'{:7.3f}'.format,'Kr':'{:5.3f}'.format,
                'Ke':'{:5.3f}'.format,'E':'{:6.3f}'.format,
                'DPe':'{:7.3f}'.format,'Kc':'{:5.3f}'.format,
                'ETc':'{:6.3f}'.format,'TAW':'{:7.3f}'.format,
                'TAWrmax':'{:7.3f}'.format,'TAWb':'{:7.3f}'.format,
                'Zr':'{:5.3f}'.format,'p':'{:5.3f}'.format,
                'RAW':'{:7.3f}'.format,'Ks':'{:5.3f}'.format,
                'Ka':'{:5.3f}'.format,'ETa':'{:6.3f}'.format,
                'T':'{:6.3f}'.format,'DP':'{:7.3f}'.format,
                'Dinc':'{:7.3f}'.format,'Dr':'{:7.3f}'.format,
                'fDr':'{:7.3f}'.format,'Drmax':'{:7.3f}'.format,
                'fDrmax':'{:7.3f}'.format,'Db':'{:7.3f}'.format,
                'fDb':'{:7.3f}'.format,'Ksend':'{:5.3f}'.format,
                'Irrig':'{:7.3f}'.format,'IrrLoss':'{:7.3f}'.format,
                'Rain':'{:7.3f}'.format,'Runoff':'{:7.3f}'.format}
        ast='*'*72
        s = ('{:s}\n'
             'pyfao56: FAO-56 Evapotranspiration in Python '
             '(ver. {:s})\n'
             'Output Data\n'
             'Timestamp: {:s}\n'
             'Simulation start date: {:s}\n'
             'Simulation end date: {:s}\n'
             'Soil method: {:s}\n'
             '{:s}\n'
             '{:s}\n'
             '{:s}\n'
             'Year-DOY  Year  DOY  DOW      Date  ETref   Kcm   ETcm'
             '  tKcb   Kcb   ETcb     h Kcmax  ETmax    fc    fw   few'
             '      De    Kr    Ke      E     DPe    Kc    ETc     TAW'
             ' TAWrmax    TAWb    Zr     p     RAW    Ks    Ka    ETa'
             '      T      DP    Dinc      Dr     fDr   Drmax  fDrmax'
             '      Db     fDb Ksend   Irrig IrrLoss    Rain  Runoff'
             '  Year  DOY  DOW      Date\n'
             ).format(ast,
                      __version__,
                      timestamp,
                      sdate,
                      edate,
                      solmthd,
                      ast,
                      self.comment,
                      ast)
        if not self.odata.empty:
            s += self.odata.to_string(header=False,formatters=fmts)
        return s

    def savefile(self,filepath='pyfao56.out'):
        """Save pyfao56 output data to a file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.out')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'w')
        except FileNotFoundError:
            print('The filepath for output data is not found.')
        else:
            f.write(self.__str__())
            f.close()

    def savesums(self, filepath='pyfao56.sum'):
        """Save a summary file with cumulative water balance values.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.sum')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        self.tmstmp = datetime.datetime.now()
        timestamp = self.tmstmp.strftime('%m/%d/%Y %H:%M:%S')
        sdate = self.startDate.strftime('%m/%d/%Y')
        edate = self.endDate.strftime('%m/%d/%Y')

        ast = '*'*72
        s = ('{:s}\n'
            'pyfao56: FAO-56 Evapotranspiration in Python\n'
            'Seasonal Water Balance Summary\n'
            'Timestamp: {:s}\n'
            'Simulation start date: {:s}\n'
            'Simulation end date: {:s}\n'
            'All values expressed in mm.\n'
            '{:s}\n'
            '{:s}\n'
            '{:s}\n'
            ).format(ast,timestamp,sdate,edate,ast,self.comment,ast)
        if not self.odata.empty:
            keys = ['ETref','ETcm','ETcb','ETmax','ETc','ETa','E','T',
                    'DP','Irrig','IrrLoss','Rain','Runoff','Dr_ini',
                    'Dr_end','Drmax_ini','Drmax_end']
            for key in keys:
                s += '{:8.3f} : {:s}\n'.format(self.swbdata[key],key)

        try:
            f = open(filepath, 'w')
        except FileNotFoundError:
            print('The filepath for summary data is not found.')
        else:
            f.write(s)
            f.close()

    class ModelState:
        """Contain parameters and states for a single timestep."""

        pass

    def run(self):
        """Initialize model, conduct simulations, update self.odata"""

        tcurrent = self.startDate
        tdelta = datetime.timedelta(days=1)

        #Initialize model state
        io = self.ModelState()
        io.i = 0
        io.Kcmini  = self.par.Kcmini
        io.Kcmmid  = self.par.Kcmmid
        io.Kcmend  = self.par.Kcmend
        io.Kcbini  = self.par.Kcbini
        io.Kcbmid  = self.par.Kcbmid
        io.Kcbend  = self.par.Kcbend
        io.Lini    = self.par.Lini
        io.Ldev    = self.par.Ldev
        io.Lmid    = self.par.Lmid
        io.Lend    = self.par.Lend
        io.hini    = self.par.hini
        io.hmax    = self.par.hmax
        io.thetaFC = self.par.thetaFC
        io.thetaWP = self.par.thetaWP
        io.theta0  = self.par.theta0
        io.Zrini   = self.par.Zrini
        io.Zrmax   = self.par.Zrmax
        io.pbase   = self.par.pbase
        io.Ze      = self.par.Ze
        io.REW     = self.par.REW
        io.CN2     = float(self.par.CN2)
        if self.sol is None:
            io.solmthd = 'D' #Default homogeneous soil from Parameters
            #Total evaporable water (TEW, mm) - FAO-56 Eq. 73
            io.TEW = 1000. * (io.thetaFC - 0.50 * io.thetaWP) * io.Ze
            #Initial depth of evaporation (De, mm) - FAO-56 page 153
            io.De = 1000. * (io.thetaFC - 0.50 * io.thetaWP) * io.Ze
            #Initial root zone depletion (Dr, mm) - FAO-56 Eq. 87
            io.Dr = 1000. * (io.thetaFC - io.theta0) * io.Zrini
            #Initial soil depletion for max root depth (Drmax, mm)
            io.Drmax = 1000. * (io.thetaFC - io.theta0) * io.Zrmax
            #Initial root zone total available water (TAW, mm)
            io.TAW = 1000. * (io.thetaFC - io.thetaWP) * io.Zrini
            #By default, FAO-56 doesn't consider the following variables
            io.TAWrmax = -99.999
            io.Db = -99.999
            io.TAWb = -99.999
        else:
            io.solmthd = 'L' #Layered soil profile from SoilProfile
            io.lyr_dpths = list(self.sol.sdata.index)
            if io.lyr_dpths[-1]*10 < io.Zrmax*1000.:
                raise Exception("Soil profile depth must be >= Zrmax")
            io.lyr_thFC  = list(self.sol.sdata['thetaFC'])
            io.lyr_thWP  = list(self.sol.sdata['thetaWP'])
            io.lyr_th0   = list(self.sol.sdata['theta0'])
            io.TEW = 0.
            io.De = 0.
            io.Dr = 0.
            io.Drmax = 0.
            io.TAW = 0.
            io.TAWrmax = 0.
            #Iterate down the soil profile in 1 mm increments
            for dpthmm in list(range(1, (io.lyr_dpths[-1] * 10 + 1))):
                #Find soil layer index that contains dpthmm
                lyr_idx = [idx for (idx, dpth) in
                          enumerate(io.lyr_dpths) if dpthmm<=dpth*10][0]
                #Total evaporable water (TEW, mm) - FAO-56 Eq. 73
                if dpthmm <= io.Ze * 1000.: #mm
                    diff=io.lyr_thFC[lyr_idx]-0.50*io.lyr_thWP[lyr_idx]
                    io.TEW += diff #mm
                #Initial depth of evaporation (De, mm) - FAO-56 page 153
                if dpthmm <= io.Ze * 1000.: #mm
                    diff=io.lyr_thFC[lyr_idx]-0.50*io.lyr_thWP[lyr_idx]
                    io.De += diff #mm
                #Initial root zone depletion (Dr, mm)
                if dpthmm <= io.Zrini * 1000.: #mm
                    diff = (io.lyr_thFC[lyr_idx] - io.lyr_th0[lyr_idx])
                    io.Dr += diff #mm
                #Initial depletion for max root depth (Drmax, mm)
                if dpthmm <= io.Zrmax * 1000.: #mm
                    diff = (io.lyr_thFC[lyr_idx] - io.lyr_th0[lyr_idx])
                    io.Drmax += diff #mm
                #Initial root zone total available water (TAW, mm)
                if dpthmm <= io.Zrini * 1000.: #mm
                    diff = (io.lyr_thFC[lyr_idx] - io.lyr_thWP[lyr_idx])
                    io.TAW += diff #mm
                #Total available water for max root depth (TAWrmax, mm)
                if dpthmm <= io.Zrmax * 1000.: #mm
                    diff = (io.lyr_thFC[lyr_idx] - io.lyr_thWP[lyr_idx])
                    io.TAWrmax += diff #mm
            #Initial depletion in the bottom layer (Db, mm)
            io.Db = io.Drmax - io.Dr
            #Initial total available water in bottom layer (TAWb, mm)
            io.TAWb = io.TAWrmax - io.TAW
        #Initial root zone soil water depletion fraction (fDr, mm/mm)
        io.fDr = 1.0 - ((io.TAW - io.Dr) / io.TAW)
        io.h = io.hini
        io.Zr = io.Zrini
        io.fw = 1.0
        io.wndht  = self.wth.wndht
        io.rfcrp  = self.wth.rfcrp
        io.roff   = self.roff
        io.cons_p = self.cons_p
        io.aq_Ks  = self.aq_Ks
        #Initial stress coefficients (Ks & Ksend, 0.0-1.0)
        #Transpiration reduction factor
        io.Ks = 1.0
        if io.aq_Ks is True:
            #Ks method from AquaCrop
            rSWD = io.Dr/io.TAW
            Drel = (rSWD-io.pbase)/(1.0-io.pbase)
            sf = 1.5
            aqKs = 1.0-(math.exp(sf*Drel)-1.0)/(math.exp(sf)-1.0)
            io.Ksend = sorted([0.0, aqKs, 1.0])[1]
        else:
            #FAO-56 Eq. 84
            io.RAW = io.pbase * io.TAW
            io.Ksend=sorted([0.0,(io.TAW-io.Dr)/(io.TAW-io.RAW),1.0])[1]
        io.Ka = io.Kcmini
        self.odata = pd.DataFrame(columns=self.cnames)

        #Adjustments of Kcmmid (FAO-56 Eq. 62), Kcmend (FAO-56 Eq. 65),
        #and Kcbmid and Kcbend (FAO-56 Eq. 70) based on minimum relative
        #humidity and wind speed
        if self.K_adj is True:
            days1 = io.Lini+io.Ldev #mid first
            days2 = io.Lini+io.Ldev+io.Lmid-1 #mid last
            days3 = io.Lini+io.Ldev+io.Lmid #end first
            days4 = io.Lini+io.Ldev+io.Lmid+io.Lend-1 #end last
            date1 = (self.startDate + days1*tdelta).strftime('%Y-%j')
            date2 = (self.startDate + days2*tdelta).strftime('%Y-%j')
            date3 = (self.startDate + days3*tdelta).strftime('%Y-%j')
            date4 = (self.startDate + days4*tdelta).strftime('%Y-%j')

            convert = 4.87/math.log(67.8*io.wndht-5.42) #Wndsp to u2

            #Compute adjustment for Kcmmid and Kcbmid
            rhmin_sub = self.wth.wdata.loc[date1:date2]['RHmin']
            rhmin_mean = sorted([20.0,rhmin_sub.mean(),80.0])[1]
            wndsp_sub = self.wth.wdata.loc[date1:date2]['Wndsp']
            u2_mean = wndsp_sub.apply(lambda x: x*convert).mean()
            u2_mean = sorted([1.0,u2_mean,6.0])[1]
            term1 = 0.04*(u2_mean-2.)
            term2 = 0.004*(rhmin_mean-45.)
            term3 = (term1-term2)*(io.hmax/3.)**0.3
            if io.Kcmmid >= 0.45 and not math.isnan(term3):
                io.Kcmmid = io.Kcmmid + round(term3,3)
            if io.Kcbmid >= 0.45 and not math.isnan(term3):
                io.Kcbmid = io.Kcbmid + round(term3,3)

            #Compute adjustment for Kcmend and Kcbend
            rhmin_sub = self.wth.wdata.loc[date3:date4]['RHmin']
            rhmin_mean = sorted([20.0,rhmin_sub.mean(),80.0])[1]
            wndsp_sub = self.wth.wdata.loc[date3:date4]['Wndsp']
            u2_mean = wndsp_sub.apply(lambda x: x*convert).mean()
            u2_mean = sorted([1.0,u2_mean,6.0])[1]
            term1 = 0.04*(u2_mean-2.)
            term2 = 0.004*(rhmin_mean-45.)
            term3 = (term1-term2)*(io.hmax/3.)**0.3
            if io.Kcmend >= 0.45 and not math.isnan(term3):
                io.Kcmend = io.Kcmend + round(term3,3)
            if io.Kcbend >= 0.45 and not math.isnan(term3):
                io.Kcbend = io.Kcbend + round(term3,3)

        while tcurrent <= self.endDate:
            mykey = tcurrent.strftime('%Y-%j')

            #Update ModelState object
            io.ETref = self.wth.wdata.loc[mykey,'ETref']
            if math.isnan(io.ETref):
                io.ETref = self.wth.compute_etref(mykey)
            io.rain = self.wth.wdata.loc[mykey,'Rain']
            io.wndsp = self.wth.wdata.loc[mykey,'Wndsp']
            if math.isnan(io.wndsp):
                io.wndsp = 2.0
            io.rhmin = self.wth.wdata.loc[mykey,'RHmin']
            if math.isnan(io.rhmin):
                tmax = self.wth.wdata.loc[mykey,'Tmax']
                tmin = self.wth.wdata.loc[mykey,'Tmin']
                tdew = self.wth.wdata.loc[mykey,'Tdew']
                if math.isnan(tdew):
                    tdew = tmin
                #ASCE (2005) Eqs. 7 and 8
                emax = 0.6108*math.exp((17.27*tmax)/(tmax+237.3))
                ea   = 0.6108*math.exp((17.27*tdew)/(tdew+237.3))
                io.rhmin = ea/emax*100.
            if math.isnan(io.rhmin):
                io.rhmin = 45.
            io.idep = 0.0
            io.ieff = 100.0
            if self.irr is not None:
                if mykey in self.irr.idata.index:
                    io.idep = self.irr.idata.loc[mykey,'Depth']
                    io.fw = self.irr.idata.loc[mykey,'fw']
                    io.ieff = self.irr.idata.loc[mykey,'ieff']

            #Evaluate autoirrigation conditions and compute amounts
            if self.autoirr is not None:
                for i in range(self.autoirr.aidata.shape[0]):
                    #Evaluate date range condition
                    aistart= self.autoirr.aidata.loc[i,'start']
                    aistart= datetime.datetime.strptime(aistart,'%Y-%j')
                    aiend  = self.autoirr.aidata.loc[i,'end']
                    aiend  = datetime.datetime.strptime(aiend,'%Y-%j')
                    if tcurrent<aistart or tcurrent>aiend:
                        continue
                    #Evaluate "after last recorded irrigation" condition
                    if self.autoirr.aidata.loc[i,'alre']:
                        if self.irr is not None:
                            lastirr = self.irr.getlastdate()
                            if tcurrent <= lastirr:
                                continue
                    #Evaluate day of the week condition
                    dnow = tcurrent.strftime('%w')
                    if dnow not in self.autoirr.aidata.loc[i,'idow']:
                        continue
                    #Evaluate forecasted precipitation condition
                    fpdep = self.autoirr.aidata.loc[i,'fpdep']
                    fpday = int(self.autoirr.aidata.loc[i,'fpday'])
                    fpact = self.autoirr.aidata.loc[i,'fpact']
                    fcrain = 0.
                    for j in range(fpday):
                        fpdate = tcurrent + j*tdelta
                        fpkey = fpdate.strftime('%Y-%j')
                        fcrain += self.wth.wdata.loc[fpkey,'Rain']
                    reduceirr = 0.
                    if fcrain >= fpdep:
                        if fpact == 'cancel':
                            continue
                        elif fpact == 'reduce':
                            reduceirr = fcrain
                        elif fpact not in ['proceed']:
                            continue
                    #Evaluate management allowed depletion (mm/mm)
                    if io.fDr <= self.autoirr.aidata.loc[i,'mad']:
                        continue
                    #Evaluate management allowed depletion (mm)
                    if io.Dr <= self.autoirr.aidata.loc[i,'madDr']:
                        continue
                    #Evaluate critical Ks
                    if io.Ksend >= self.autoirr.aidata.loc[i,'ksc']:
                        continue
                    #Evaluate days since last irrigation (dsli)
                    idays = self.odata[self.odata['Irrig']>0.]
                    idays = pd.to_datetime(idays.index,format='%Y-%j')
                    if idays.size > 0:
                        dsli = (tcurrent-max(idays)).days
                    else:
                        dsli = ((tcurrent-self.startDate).days)+1
                    if dsli < self.autoirr.aidata.loc[i,'dsli']:
                        continue
                    #Evaluate days since last watering event
                    evnt = self.autoirr.aidata.loc[i,'evnt']
                    edays = self.odata[(self.odata['Irrig']-
                                        self.odata['IrrLoss']+
                                        self.odata['Rain']-
                                        self.odata['Runoff'])>=evnt]
                    edays = pd.to_datetime(edays.index,format='%Y-%j')
                    if edays.size > 0:
                        dsle = (tcurrent-max(edays)).days
                    else:
                        dsle = ((tcurrent-self.startDate).days)+1
                    if dsle < self.autoirr.aidata.loc[i,'dsle']:
                        continue

                    #All conditions were met, need to autoirrigate.
                    #Target a full profile (Dr=0) at end of timestep.
                    #Use yesterday's Dr (io.Dr) plus an estimate of
                    #today's ETa (from today's ETref and yesterday's Ka)
                    ETest = io.Ka * io.ETref
                    rate = max([0.0,io.Dr + ETest - reduceirr])

                    #Alternatively, the default rate may be modified:
                    #Use a fixed rate
                    ifix  = self.autoirr.aidata.loc[i,'ifix']
                    if not math.isnan(ifix):
                        rate = max([0.0, ifix - reduceirr])
                    #Target a specific root-zone soil water depletion
                    itdr  = self.autoirr.aidata.loc[i,'itdr']
                    if not math.isnan(itdr):
                        rate = max([0.0,io.Dr+ETest-reduceirr-itdr])
                    #Target a fractional root-zone soil water depletion
                    itfdr = self.autoirr.aidata.loc[i,'itfdr']
                    if not math.isnan(itfdr):
                        itdr2 = io.TAW-io.TAW*(1.0-itfdr)
                        rate = max([0.0,io.Dr+ETest-reduceirr-itdr2])
                    #Use ETa less precip for past X number of days
                    ettyp = self.autoirr.aidata.loc[i,'ettyp']
                    ietrd = self.autoirr.aidata.loc[i,'ietrd']
                    if not math.isnan(ietrd):
                        dsss = (tcurrent-self.startDate).days
                        recent = self.odata.tail(min([dsss,int(ietrd)]))
                        p1 = recent['Rain'].sum()
                        p2 = recent['Runoff'].sum()
                        et = recent[ettyp].sum()
                        etrd=(et-p1+p2)
                        rate = max([0.0,etrd - reduceirr])
                    #Use ETa less precip since last irrigation
                    ettyp = self.autoirr.aidata.loc[i,'ettyp']
                    ietri = self.autoirr.aidata.loc[i,'ietri']
                    if ietri:
                        dsss = (tcurrent-self.startDate).days
                        recent = self.odata.tail(min([dsss,dsli]))
                        p1 = recent['Rain'].sum()
                        p2 = recent['Runoff'].sum()
                        et = recent[ettyp].sum()
                        etri=(et-p1+p2)
                        rate = max([0.0,etri - reduceirr])
                    #Use ETa less precip since last watering event
                    ettyp = self.autoirr.aidata.loc[i,'ettyp']
                    ietre = self.autoirr.aidata.loc[i,'ietre']
                    if ietre:
                        dsss = (tcurrent-self.startDate).days
                        recent = self.odata.tail(min([dsss,dsle]))
                        p1 = recent['Rain'].sum()
                        p2 = recent['Runoff'].sum()
                        et = recent[ettyp].sum()
                        etre=(et-p1+p2)
                        rate = max([0.0,etre - reduceirr])

                    #Furthermore, adjustments to the rate can be made
                    #Adjust rate by a fixed percentage
                    iper  = self.autoirr.aidata.loc[i,'iper']
                    if not math.isnan(iper):
                        rate = max([0.0, rate*iper/100.])
                    #Adjust rate for irrigation inefficiency
                    ieff  = self.autoirr.aidata.loc[i,'ieff']
                    if not math.isnan(ieff):
                        rate = rate/(ieff/100.)
                        io.ieff = ieff
                    #Adjust rate for minimum irrigation amount
                    imin  = self.autoirr.aidata.loc[i,'imin']
                    if not math.isnan(imin):
                        rate = max([imin, rate])
                    #Adjust rate for maximum irrigation amount
                    imax  = self.autoirr.aidata.loc[i,'imax']
                    if not math.isnan(imax):
                        rate = min([imax,rate])

                    #Update fraction wetted (fw) for autoirrigation
                    io.fw=self.autoirr.aidata.loc[i,'fw']

                    #Specify the final autoirrigation rate
                    io.idep=rate
                    break

            #Obtain updates for Kcb, h, and fc, if available
            io.updKcb = float('NaN')
            io.updh = float('NaN')
            io.updfc = float('NaN')
            if self.upd is not None:
                io.updKcb = self.upd.getdata(mykey,'Kcb')
                io.updh = self.upd.getdata(mykey,'h')
                io.updfc = self.upd.getdata(mykey,'fc')

            #Advance timestep
            self._advance(io)

            #Append results to self.odata
            year = tcurrent.strftime('%Y')
            doy = tcurrent.strftime('%j') #Day of Year
            dow = tcurrent.strftime('%a') #Day of Week
            dat = tcurrent.strftime('%m/%d/%y') #Date mm/dd/yy
            data = [year, doy, dow, dat, io.ETref, io.Kcm, io.ETcm,
                    io.tKcb, io.Kcb, io.ETcb, io.h, io.Kcmax, io.ETmax,
                    io.fc, io.fw, io.few, io.De, io.Kr, io.Ke, io.E,
                    io.DPe, io.Kc, io.ETc, io.TAW, io.TAWrmax, io.TAWb,
                    io.Zr, io.p, io.RAW, io.Ks, io.Ka, io.ETa, io.T,
                    io.DP, io.Dinc, io.Dr, io.fDr, io.Drmax, io.fDrmax,
                    io.Db, io.fDb, io.Ksend, io.idep, io.irrloss,
                    io.rain, io.runoff, year, doy, dow, dat]
            self.odata.loc[mykey] = data

            tcurrent = tcurrent + tdelta
            io.i+=1

        #Save seasonal water balance data to self.swbdata dictionary
        sdoy = self.startDate.strftime("%Y-%j")
        edoy = self.endDate.strftime("%Y-%j")
        self.swbdata = {'ETref'    :sum(self.odata['ETref']),
                        'ETcm'     :sum(self.odata['ETcm']),
                        'ETcb'     :sum(self.odata['ETcb']),
                        'ETmax'    :sum(self.odata['ETmax']),
                        'ETc'      :sum(self.odata['ETc']),
                        'ETa'      :sum(self.odata['ETa']),
                        'E'        :sum(self.odata['E']),
                        'T'        :sum(self.odata['T']),
                        'DP'       :sum(self.odata['DP']),
                        'Irrig'    :sum(self.odata['Irrig']),
                        'IrrLoss'  :sum(self.odata['IrrLoss']),
                        'Rain'     :sum(self.odata['Rain']),
                        'Runoff'   :sum(self.odata['Runoff']),
                        'Dr_ini'   :self.odata.loc[sdoy,'Dr'],
                        'Dr_end'   :self.odata.loc[edoy,'Dr'],
                        'Drmax_ini':self.odata.loc[sdoy,'Drmax'],
                        'Drmax_end':self.odata.loc[edoy,'Drmax']}

    def _advance(self, io):
        """Advance the model by one daily timestep.

        Parameters
        ----------
        io : ModelState object
        """

        #Trapezoidal basal crop coefficient (tKcb)
        #Basal crop coefficient (Kcb) considering updates
        #Single (mean) crop coeffient (Kcm)
        #From FAO-56 Table 11 for growth stage lengths,
        #Table 12 for Kcm and Table 17 for tKcb and Kcb
        s1 = io.Lini
        s2 = s1 + io.Ldev
        s3 = s2 + io.Lmid
        s4 = s3 + io.Lend
        if 0<=io.i<=s1:
            io.tKcb = io.Kcbini
            io.Kcb = io.Kcbini
            io.Kcm = io.Kcmini
        elif s1<io.i<=s2:
            io.tKcb += (io.Kcbmid-io.Kcbini)/(s2-s1)
            io.Kcb += (io.Kcbmid-io.Kcbini)/(s2-s1)
            io.Kcm += (io.Kcmmid-io.Kcmini)/(s2-s1)
        elif s2<io.i<=s3:
            io.tKcb = io.Kcbmid
            io.Kcb = io.Kcbmid
            io.Kcm = io.Kcmmid
        elif s3<io.i<=s4:
            io.tKcb += (io.Kcbmid-io.Kcbend)/(s3-s4)
            io.Kcb += (io.Kcbmid-io.Kcbend)/(s3-s4)
            io.Kcm += (io.Kcmmid-io.Kcmend)/(s3-s4)
        elif s4<io.i:
            io.tKcb = io.Kcbend
            io.Kcb = io.Kcbend
            io.Kcm = io.Kcmend

        #Crop evapotranspiration, single method (ETcm) - FAO-56 Eq. 56
        io.ETcm = io.Kcm * io.ETref

        #Basal evapotranspiration (ETcb) - DeJonge et al. (2025) Table 1
        io.ETcb = io.Kcb * io.ETref

        #Overwrite Kcb if updates are available
        if io.updKcb > 0: io.Kcb = io.updKcb

        #Plant height (h, m)
        io.h = max([io.hini+(io.hmax-io.hini)*(io.Kcb-io.Kcbini)/
                    (io.Kcbmid-io.Kcbini),0.001,io.h])
        #Overwrite h if updates are available
        if io.updh > 0: io.h = io.updh

        #Root depth (Zr, m) - FAO-56 page 279
        io.Zr = max([io.Zrini + (io.Zrmax-io.Zrini)*(io.tKcb-io.Kcbini)/
                     (io.Kcbmid-io.Kcbini),0.001,io.Zr])

        #Maximum or upper limit crop coefficient (Kcmax) - FAO-56 Eq. 72
        u2 = io.wndsp * (4.87/math.log(67.8*io.wndht-5.42))
        u2 = sorted([1.0,u2,6.0])[1]
        rhmin = sorted([20.0,io.rhmin,80.0])[1]
        if io.rfcrp == 'S':
            io.Kcmax = max([1.2+(0.04*(u2-2.0)-0.004*(rhmin-45.0))*
                            (io.h/3.0)**.3, io.Kcb+0.05])
        elif io.rfcrp == 'T':
            io.Kcmax = max([1.0, io.Kcb + 0.05])

        #Maximum evapotranspiration (ETmax) - DeJonge et al. (2025) Tbl1
        io.ETmax = io.Kcmax * io.ETref

        #Canopy cover fraction (fc, 0.0-0.99) - FAO-56 Eq. 76
        io.fc = sorted([0.0,((io.Kcb-io.Kcbini)/(io.Kcmax-io.Kcbini))**
                        (1.0+0.5*io.h),0.99])[1]
        #Overwrite fc if updates are available
        if io.updfc > 0: io.fc = io.updfc

        #Losses due to irrigation inefficiency (irrloss, mm)
        io.irrloss = io.idep - io.idep * (io.ieff / 100.)

        #Effective irrigation (mm)
        effirr = io.idep - io.irrloss

        #Surface runoff (runoff, mm)
        io.runoff = 0.0
        if io.roff is True:
            #Method per ASCE (2016) Eqs. 14-12 to 14-20, page 451-454
            CN1 = io.CN2/(2.281-0.01281*io.CN2) #ASCE (2016) Eq. 14-14
            CN3 = io.CN2/(0.427+0.00573*io.CN2) #ASCE (2016) Eq. 14-15
            if io.De <= 0.5*io.REW:
                CN = CN3 #ASCE (2016) Eq. 14-18
            elif io.De >= 0.7*io.REW+0.3*io.TEW:
                CN = CN1 #ASCE (2016) Eq. 14-19
            else:
                CN = (io.De-0.5*io.REW)*CN1
                CN = CN+(0.7*io.REW+0.3*io.TEW-io.De)*CN3
                CN = CN/(0.2*io.REW+0.3*io.TEW) #ASCE (2016) Eq. 14-20
            storage = 250.*((100./CN)-1.) #ASCE (2016) Eq. 14-12
            if io.rain > 0.2*storage:
                #ASCE (2016) Eq. 14-13
                io.runoff = (io.rain-0.2*storage)**2
                io.runoff = io.runoff/(io.rain+0.8*storage)
                io.runoff = min([io.runoff,io.rain])
            else:
                io.runoff = 0.0

        #Effective precipitation (mm)
        effrain = io.rain - io.runoff

        #Fraction soil surface wetted (fw) - FAO-56 Table 20, page 149
        if io.idep > 0.0 and io.rain > 0.0:
            pass #fw=fw input
        elif io.idep > 0.0 and io.rain <= 0.0:
            pass #fw=fw input
        elif io.idep <= 0.0 and io.rain >= 3.0:
            io.fw = 1.0
        else:
            pass #fw = previous fw

        #Exposed & wetted soil fraction (few, 0.01-1.0) - FAO-56 Eq. 75
        io.few = sorted([0.01,min([1.0-io.fc, io.fw]),1.0])[1]

        #Evaporation reduction coefficient (Kr, 0-1) - FAO-56 Eq. 74
        io.Kr = sorted([0.0,(io.TEW-io.De)/(io.TEW-io.REW),1.0])[1]

        #Evaporation coefficient (Ke) - FAO-56 Eq. 71
        io.Ke = min([io.Kr*(io.Kcmax-io.Kcb), io.few*io.Kcmax])

        #Soil water evaporation (E, mm) - FAO-56 Eq. 69
        io.E = io.Ke * io.ETref

        #Deep percolation under exposed soil (DPe, mm) - FAO-56 Eq. 79
        io.DPe = max([effrain + effirr/io.fw - io.De, 0.0])

        #Cumulative depth of evaporation (De, mm) - FAO-56 Eqs. 77 & 78
        De = io.De - effrain - effirr/io.fw + io.E/io.few + io.DPe
        io.De = sorted([0.0,De,io.TEW])[1]

        #Dual drop coefficient, standard conditions (Kc) - FAO-56 Eq. 69
        io.Kc = io.Ke + io.Kcb

        #Non-stressed evapotranspiration (ETc, mm) - FAO-56 Eq. 69
        io.ETc = io.Kc * io.ETref

        if io.solmthd == 'D':
            # Total available water (TAW, mm) - FAO-56 Eq. 82
            io.TAW = 1000.0 * (io.thetaFC - io.thetaWP) * io.Zr
        elif io.solmthd == 'L':
            io.TAW = 0.
            #Iterate down the soil profile in 1 mm increments
            for dpthmm in list(range(1, (io.lyr_dpths[-1] * 10 + 1))):
                #Find soil layer index that contains dpthmm
                lyr_idx = [idx for (idx, dpth) in
                          enumerate(io.lyr_dpths) if dpthmm<=dpth*10][0]
                #Total available water (TAW, mm)
                if dpthmm <= io.Zr * 1000.: #mm
                    diff = (io.lyr_thFC[lyr_idx] - io.lyr_thWP[lyr_idx])
                    io.TAW += diff #mm
            #Total available water in the bottom layer (TAWb, mm)
            io.TAWb_prev = io.TAWb
            io.TAWb = io.TAWrmax - io.TAW

        #Fraction depleted TAW (p, 0.1-0.8) - FAO-56 p162 and Table 22
        if io.cons_p is True:
            io.p = io.pbase
        else:
            io.p = sorted([0.1,io.pbase+0.04*(5.0-io.ETc),0.8])[1]

        #Readily available water (RAW, mm) - FAO-56 Equation 83
        io.RAW = io.p * io.TAW

        #Stress coefficient at beginning of timestep (Ks,0.0-1.0)
        #Transpiration reduction factor
        if io.aq_Ks is True:
            #Ks method from AquaCrop
            rSWD = io.Dr/io.TAW
            Drel = (rSWD-io.p)/(1.0-io.p)
            sf = 1.5
            aqKs = 1.0-(math.exp(sf*Drel)-1.0)/(math.exp(sf)-1.0)
            io.Ks = sorted([0.0, aqKs, 1.0])[1]
        else:
            #FAO-56 Eq. 84
            io.Ks = sorted([0.0,(io.TAW-io.Dr)/(io.TAW-io.RAW),1.0])[1]

        #Actual crop coefficient (Ka) - FAO-56 Eq. 80
        #DeJonge et al. (2025) updates terminology: Kcadj to Ka
        io.Ka = io.Ks * io.Kcb + io.Ke

        #Actual evapotranspiration (ETa, mm) - FAO-56 Eq. 80
        #DeJonge et al. (2025) updates terminology: ETcadj to ETa
        io.ETa = io.Ka * io.ETref

        #Actual plant transpiration (T, mm)
        io.T = (io.Ks * io.Kcb) * io.ETref

        #Water balance methods
        if io.solmthd == 'D':
            #Deep percolation (DP, mm) - FAO-56 Eq. 88
            #Boundary layer is considered at the root zone depth (Zr)
            DP = effrain + effirr - io.ETa - io.Dr
            io.DP = max([DP,0.0])

            #Root zone soil water depletion (Dr,mm) - FAO-56 Eqs.85 & 86
            Dr = io.Dr - effrain - effirr + io.ETa + io.DP
            io.Dr = sorted([0.0, Dr, io.TAW])[1]

            #Root zone soil water depletion fraction (fDr, mm/mm)
            io.fDr = 1.0 - ((io.TAW - io.Dr) / io.TAW)

            #By default, FAO-56 doesn't consider the following variables
            io.Dinc = -99.999
            io.Drmax = -99.999
            io.fDrmax = -99.999
            io.Db = -99.999
            io.fDb = -99.999

        elif io.solmthd == 'L':
            #Deep percolation (DP, mm)
            #Boundary layer is at the max root depth (Zrmax)
            DP = effrain + effirr - io.ETa - io.Drmax
            io.DP = max([DP,0.0])

            #Depletion increment due to root growth (Dinc, mm)
            #Computed from Db based on the incremental change in TAWb
            if io.TAWb_prev > 0.0:
                io.Dinc = io.Db * (1.0 - (io.TAWb / io.TAWb_prev))
            else: #handle zero divide issue
                io.Dinc = 0.0

            #Root zone soil water depletion (Dr, mm)
            Dr = io.Dr - effrain - effirr + io.ETa + io.Dinc
            io.Dr = sorted([0.0, Dr, io.TAW])[1]

            #Root zone soil water depletion fraction (fDr, mm/mm)
            io.fDr = 1.0 - ((io.TAW - io.Dr) / io.TAW)

            #Soil water depletion at max root depth (Drmax, mm)
            Drmax = io.Drmax - effrain - effirr + io.ETa + io.DP
            io.Drmax = sorted([0.0, Drmax, io.TAWrmax])[1]

            #Soil water depletion fraction at Zrmax (fDrmax, mm/mm)
            io.fDrmax = 1.0 - ((io.TAWrmax - io.Drmax) / io.TAWrmax)

            #Soil water depletion in the bottom layer (Db, mm)
            Db = io.Drmax - io.Dr
            io.Db = sorted([0.0, Db, io.TAWb])[1]

            #Bottom layer soil water depletion fraction (fDb, mm/mm)
            if io.TAWb > 0.0:
                io.fDb = 1.0 - ((io.TAWb - io.Db) / io.TAWb)
            else:
                io.fDb = 0.0

        #Stress coefficient at end of timestep (Ksend,0.0-1.0)
        #Transpiration reduction factor
        if io.aq_Ks is True:
            #Ks method from AquaCrop
            rSWD = io.Dr/io.TAW
            Drel = (rSWD-io.p)/(1.0-io.p)
            sf = 1.5
            aqKs = 1.0-(math.exp(sf*Drel)-1.0)/(math.exp(sf)-1.0)
            io.Ksend = sorted([0.0, aqKs, 1.0])[1]
        else:
            #FAO-56 Eq. 84
            io.Ksend=sorted([0.0,(io.TAW-io.Dr)/(io.TAW-io.RAW),1.0])[1]



"""
########################################################################
The tools subpackage within the pyfao56 Python package provides tools
that facilitate implementation of the model, but are not required to run
the model.

The tools subpackage contains the following modules:
    forecast.py - Obtains weather forecast data from the National
        Digital Forecast Database (NDFD)
    soil_water.py  - provides I/O and computational tools for processing
        and using measured volumetric soil water content data

10/17/2022 Subpackage created by Josh Brekel
11/09/2022 Initial Visualization scripts created by Josh Brekel
03/02/2023 SoilWater Class functions developed by Josh Brekel, USDA-ARS
08/25/2023 Moved forecast.py from custom to tools
08/29/2023 Added visualization.py
########################################################################
"""

from .forecast import Forecast
from .soil_water import SoilWaterSeries
from .visualization import Visualization
from .statistics import Statistics
from .tables import FAO56Tables
from .blue_green import BlueGreen



"""
########################################################################
The forecast.py module contains the Forecast class, which is used to
retrieve seven-day weather forecasts from the National Digital Forecast
Database (NDFD). It uses the REST approach, which was more robust than
the SOAP method, in terms of server responsiveness. Data is retrieved
for computation of ASCE Standardized Reference Evapotranspiration,
including wind speed (m/s) and minimum, maximum, and dew point air
temperatures (degrees C). Solar radiation forecasts are not directly
provided by NDFD, but NDFD does provide cloud cover forecasts. By
providing the optional "elevation" parameter, users can obtain solar
radiation forecasts computed by multiplication of cloud cover and
clear-sky solar radiation. Liquid precipitation forecasts (mm) are also
obtained.

#https://graphical.weather.gov/xml/rest.php
#https://graphical.weather.gov/xml/docs/elementInputNames.php
#https://digital.weather.gov/staticpages/definitions.php

xx/xx/2019 Initial Python functions developed by Kelly Thorp
11/04/2021 Finalized updates for inclusion in the pyfao56 Python package
06/05/2023 Added Srad and Rain columns -- Josh Brekel, USDA-ARS
08/17/2023 Added Clds column to output
########################################################################
"""

import datetime
import pandas as pd
import requests
import xml.etree.ElementTree as ET
import numpy as np
import math

class Forecast():
    """A class for obtaining weather forecasts from the NDFD

    Obtains weather forecast data from the National Digital Forecast
    Database (NDFD). Given latitude and longitude, the class methods
    will obtain and store a seven-day weather forecast from today
    forward, including liquid precipitation, wind speed, cloud cover,
    and minimum, maximum, and dew point air temperatures. If the
    optional elevation parameter is provided, then daily incoming solar
    radiation is forecasted from cloud cover forecasts and clear-sky
    solar radiation.

    Attributes
    ----------
    latitude : float
        Site latitude (decimal degrees)
    longitude : float
        Site longitude (decimal degrees)
    wndht : float, optional
        Correction height for NDFD wind speed (m)
        Should be the same as wndht in Weather class
        (default = 10.0, no correction)
    elevation : float, optional
        Site elevation (m) (default = NaN)
    forecast : DataFrame
        Weather forecast data from NDFD as float
        index - Year and day of year as string ('yyyy-ddd')
        columns - ['Clds','Srad','Tmax','Tmin','Tdew','Wndsp','Rain']
            Clds  - Daily cloud cover (%)
            Srad  - Daily incoming solar radiation (MJ/m2)
            Tmax  - Daily maximum air temperature (deg C)
            Tmin  - Daily minimum air temperature (deg C)
            Tdew  - Daily average dew point temperature (deg C)
            Wndsp - Daily average wind speed (m/s)
            Rain  - Daily amount of liquid precipitation (mm)

    Methods
    -------
    getforecast()
        Request and process weather forecast data from NDFD and update
        the self.forecast DataFrame.
    """

    def __init__(self, latitude, longitude, wndht=10.0,
                 elevation=float('NaN')):
        """Initialize the Forecast class attributes

        Parameters
        ----------
        latitude : float
            Site latitude (decimal degrees)
        longitude : float
            Site longitude (decimal degrees)
        wndht : float, optional
            Correction height for NDFD wind speed (m)
            Should be the same as wndht in Weather class
            (default = 10.0, no correction)
        elevation : float, optional
            Site elevation (m) (default = NaN)
        """

        self.latitude = latitude
        self.longitude = longitude
        self.wndht = wndht
        self.elevation = elevation
        self._initialize()
        self._compute_rso()

    def __str__(self):
        """Represent the Forecast class variables as a string."""
        pd.options.display.float_format = '{:6.2f}'.format
        s = ('Latitude: {:12.7f}\n'
             'Longitude: {:12.7f}\n'
             'Elevation: {:12.7f}\n\n'
             'NDFD weather forecast data for year-doy:\n'
             ).format(self.latitude, self.longitude, self.elevation)
        s += self.forecast.to_string()
        return s

    def _initialize(self):
        """Initialize the self.forecast DataFrame."""
        init = []
        keys = []
        cols = ['Clds','Srad','Tmax','Tmin','Tdew','Wndsp','Rain']
        today = datetime.datetime.today()
        NaN = float('NaN')
        for i in list(range(-1,10)):
            day = today + datetime.timedelta(days=i)
            keys.append(day.strftime('%Y-%j'))
            init.append([NaN,NaN,NaN,NaN,NaN,NaN,NaN])
        self.forecast = pd.DataFrame(init,index=keys,columns=cols)

    def _compute_rso(self):
        """Compute clear sky solar radiation on forecast days."""
        self.rso = {}
        today = datetime.datetime.today()
        NaN = float('NaN')
        for i in list(range(-1,10)):
            day = today + datetime.timedelta(days=i)
            key = day.strftime('%Y-%j')

            #ra (float) : Extraterrestrial radiation (MJ m^-2 d^-1)
            #ASCE (2005) Eqs. 21-27
            doy = int(day.strftime('%j'))
            latrad = self.latitude*math.pi/180.0 #Eq.22
            dr = 1.0+0.033*math.cos(2.0*math.pi/365.0*doy) #Eq.23
            ldelta = 0.409*math.sin(2.0*math.pi/365.0*doy-1.39) #Eq.24
            ws = math.acos(-1.0*math.tan(latrad)*math.tan(ldelta))#Eq.27
            ra1 = ws*math.sin(latrad)*math.sin(ldelta) #Eq.21
            ra2 = math.cos(latrad)*math.cos(ldelta)*math.sin(ws) #Eq.21
            ra = 24.0/math.pi*4.92*dr*(ra1+ra2) #Eq.21

            #rso (float) : Clear sky solar radiation (MJ m^-2 d^-1)
            #ASCE (2005) Eq. 19
            if not math.isnan(self.elevation):
                self.rso.update({key:(0.75+2e-5*self.elevation)*ra})
            else:
                self.rso.update({key:NaN})

    def getforecast(self):
        """Request and process weather forecast data from NDFD."""

        #Submit the request to NDFD.
        request = {}
        request.update({'lat':self.latitude})
        request.update({'lon':self.longitude})
        request.update({'product':'time-series'})
        request.update({'begin':''})
        request.update({'end':''})
        request.update({'Unit':'m'})
        request.update({'sky': 'sky'})
        request.update({'maxt':'maxt'})
        request.update({'mint':'mint'})
        request.update({'dew':'dew'})
        request.update({'wspd':'wspd'})
        request.update({'qpf':'qpf'})
        url = 'https://graphical.weather.gov/xml/sample_products/'
        url+= 'browser_interface/ndfdXMLclient.php'
        r = requests.get(url,params=request)
        #print(r.text)

        #Process the resulting XML string from NDFD.
        tree = ET.fromstring(r.text)
        data = tree.find('data')
        pars = data.find('parameters')

        items = {'Clds' :['cloud-amount','total'   ,'start-valid-time'],
                 'Tmax' :['temperature','maximum'  ,'start-valid-time'],
                 'Tmin' :['temperature','minimum'  ,'end-valid-time'  ],
                 'Tdew' :['temperature','dew point','start-valid-time'],
                 'Wndsp':['wind-speed' ,'sustained','start-valid-time'],
                 'Rain' :['precipitation', 'liquid','start-valid-time']}

        self._initialize()
        times = []
        values = []
        dayvals = []
        today = datetime.datetime.today()
        for item in items.keys():
            values[:] = []
            tag = items[item][0]
            att = items[item][1]
            node1 = pars.find("./"+tag+"[@type='"+att+"']")
            for child in node1:
                if child.tag == 'value':
                    values.append(float(child.text))
            tl = node1.attrib['time-layout']
            times[:] = []
            node2 = data.find(".//layout-key[.='"+tl+"']..")
            for child2 in node2:
                if child2.tag == items[item][2]:
                    times.append(child2.text)
            for i in list(range(-1,10)):
                day = today + datetime.timedelta(days=i)
                key1 = day.strftime('%Y-%j')
                dayvals[:] = []
                for j in list(range(len(times))):
                    Ymd='%Y-%m-%d'
                    date = datetime.datetime.strptime(times[j][:10],Ymd)
                    key2 = date.strftime('%Y-%j')
                    if key1 == key2:
                        dayvals.append(values[j])
                if len(dayvals) > 0:
                    mean = np.mean(np.array(dayvals))
                    if item == 'Wndsp' and self.wndht != 10.0:
                        # Adjust NDFD wind speed from 10.0 m to station
                        # height (wndht in Weather class)
                        zNDFD = 10.0
                        # u2 : wind speed at 2 m (m s^-1)
                        #ASCE (2005) Eq. 33 and Appendix E
                        u2 = mean * (4.87/math.log(67.8*zNDFD-5.42))
                        # uz : wind speed at wndht in Weather (m s^-1)
                        # ASCE (2005) Eq. 33 and Appendix E
                        uz = u2 / (4.87/math.log(67.8*self.wndht-5.42))
                        mean = uz
                    self.forecast.loc[key1,item] = mean
                    #Estimate Srad from cloud cover and rso
                    if item == 'Clds':
                        srad = (1.0 - mean/100.) * self.rso[key1]
                        self.forecast.loc[key1,'Srad'] = srad



"""
########################################################################
The azmet_maricopa.py module contains the AzmetMaricopa class, which
inherits from the pyfao56 Weather class in weather.py. AzmetMaricopa
provides specific I/O functionality for obtaining required weather input
data from the Arizona Meteorological Network (AZMET) station in
Maricopa, Arizona.

01/07/2016 Initial Python functions developed by Kelly Thorp
11/04/2021 Finalized updates for inclusion in the pyfao56 Python package
########################################################################
"""

from pyfao56 import Weather
import datetime
import urllib3
import math
import pandas as pd
from pyfao56.tools import Forecast

class AzmetMaricopa(Weather):
    """A class for obtaining weather data for Maricopa, Arizona

    Obtains and prepares weather data from the Arizona Meteorological
    Network (AZMET) station in Maricopa, Arizona. If necessary, obtains
    a 7-day weather forecast for Maricopa from the National Digital
    Forecast Database (NDFD) using the Forecast class in forecast.py.
    Computes ASCE Standardized Reference Evapotranspiration for the
    resulting data set. Checks for missing weather data. The class
    inherits from the pyfao56 Weather class.

    Attributes
    ----------
    rfcrp : str
        Type of reference crop  - Short ('S') or Tall ('T')
    z : float
        Weather station elevation (z) (m)
    lat : float
        Weather station latitude (decimal degrees)
    wndht : float
        Weather station wind speed measurement height (m)
    cnames : list
        Column names for wdata
    wdata : DataFrame
        Weather data as float
        index - Year and day of year as string ('yyyy-ddd')
        columns - ['Srad','Tmax','Tmin','Tdew','RHmax','RHmin',
                   'Wndsp','Rain','ETref','MorP']
            Srad  - Incoming solar radiation (MJ/m2)
            Tmax  - Daily maximum air temperature (deg C)
            Tmin  - Daily minimum air temperature (deg C)
            Vapr  - Daily average vapor pressure (kPa)
            Tdew  - Daily average dew point temperature (deg C)
            RHmax - Daily maximum relative humidity (%)
            RHmin - Daily minimum relative humidity (%)
            Wndsp - Daily average wind speed (m/s)
            Rain  - Daily precipitation (mm)
            ETref - Daily reference ET (mm)
            MorP  - Measured ('M') or Predicted ('P') data

    Methods
    -------
    customload(start,end,usefc=True)
        Overridden method from the pyfao56 Weather class to provide
        customization for loading weather data from the Maricopa AZMET
        station and weather forecasts from the National Digital Forecast
        Database (NDFD).
    """

    def customload(self,start,end,rfcrp='S',usefc=True):
        """Prepare the wdata DataFrame with required weather data.

        Parameters
        ----------
        start : str
            Simulation start year and doy ('yyyy-ddd')
        end   : str
            Simulation end year and doy ('yyyy-ddd')
        rfcrp : str, optional
            Define the reference crop (default='S')
        usefc : bool, optional
            Use the 7-day NDFD weather forecast or not (default=True)
        """

        #Define Maricopa weather station parameters
        self.rfcrp = rfcrp   #Set the reference crop
        self.z     = 361.000 #Weather station elevation (z) (m)
        self.lat   = 33.0690 #Weather station latitude (deg)
        self.wndht = 3.00000 #Wind speed measurement height (m)

        #Retrieve AZMET weather history for Maricopa
        print('Retrieving AZMET weather history for Maricopa, AZ...')
        today = datetime.datetime.today()
        azmet = []
        for year in range(1987,int(today.strftime('%Y'))+1):
            print('Retrieving {:4d} data...'.format(year))
            year2 = ('{:4d}'.format(year))[2:4]
            client = urllib3.PoolManager()
            url = 'azmet.arizona.edu/azmet/data/06'+year2+'rd.txt'
            page = client.request('GET',url,retries=9999)
            weatherdata = page.data.decode('utf-8').split('\n')[:-1]
            for line in weatherdata:
                if line in ['']: continue
                line = line.rstrip().split(',')
                lineyear = int(line[0])
                linedoy = int(line[1])
                if lineyear < 100: lineyear+=1900
                mykey = '{:04d}-{:03d}'.format(lineyear,linedoy)
                mydict = {}
                mydict.update({'Date':mykey})
                if 0.0 <= float(line[10]) <= 110.0:
                    mydict.update({'Srad':float(line[10])})
                if 0.0 <= float(line[3]) <= 55.0:
                    mydict.update({'Tmax':float(line[3])})
                if -15.0 <= float(line[4]) <= 40.0:
                    mydict.update({'Tmin':float(line[4])})
                mydict.update({'Vapr':float('NaN')})
                if year >= 2003:
                    if -50.0 <= float(line[27]) <= 50.0:
                        mydict.update({'Tdew':float(line[27])})
                else:
                    tavg = float(line[5]) #Avg daily temperature
                    havg = float(line[8]) #Avg relative humidity
                    #From https://cals.arizona.edu/azmet/dewpoint.html
                    B = math.log(havg/100.0)+((17.27*tavg)/(237.3+tavg))
                    B = B/17.27
                    D = (237.3*B)/(1-B)
                    if -50.0 <= D <= 50.0:
                        mydict.update({'Tdew':D})
                if 0.0 <= float(line[6]) <= 100.0:
                    mydict.update({'RHmax':float(line[6])})
                if 0.0 <= float(line[7]) <= 100.0:
                    mydict.update({'RHmin':float(line[7])})
                if 0.0 <= float(line[18]) <= 30.0:
                    mydict.update({'Wndsp':float(line[18])})
                if 0.0 <= float(line[11]) <= 200.0:
                    mydict.update({'Rain':float(line[11])})
                azmet.append(mydict)
        azmet = pd.DataFrame(azmet)
        lessvapr = azmet.drop('Vapr', axis=1)
        nanrows = lessvapr.isna().any(1).to_numpy().nonzero()[0]
        for item in nanrows:
            mykey = azmet.loc[item,'Date']
            print('Warning: Questionable weather data: ' + mykey)
        if len(nanrows) > 0:
            result = input('Continue (Y/N)?')
            if result not in ['Y']:
                return

        #Process AZMET data from requested start to end date
        startDate = datetime.datetime.strptime(start, '%Y-%j')
        endDate   = datetime.datetime.strptime(end, '%Y-%j')
        tdelta = datetime.timedelta(days=1)
        yesterday = today - tdelta
        tcurrent = startDate
        future = []
        wthdata = []
        needfuture = False
        print('Processing AZMET weather data...')
        while tcurrent <= endDate:
            mykey = tcurrent.strftime('%Y-%j')
            if tcurrent <= yesterday: #Get data for days prior to today
                daydata = azmet.loc[azmet['Date'] == mykey]
                daydata = daydata.to_dict('records')[0]
                daydata.update({'MorP':'M'})
                wthdata.append(daydata)
            else: #Predict future data as average of past data on a day
                needfuture = True
                future[:] = []
                for year in range(1987,int(tcurrent.strftime('%Y'))):
                    mon = int(tcurrent.strftime('%m')) #month
                    dom = int(tcurrent.strftime('%d')) #day of month
                    if mon==2 and dom==29: #leap day
                        if year%4: #not leapyear, use feb 28 data
                            feb28=tcurrent-tdelta
                        pastdate = feb28.replace(year=year)
                    else:
                        pastdate = tcurrent.replace(year=year)
                    pastkey = pastdate.strftime('%Y-%j')
                    daydata = azmet.loc[azmet['Date'] == pastkey]
                    daydata = daydata.to_dict('records')[0]
                    future.append(daydata)
                futmean = pd.DataFrame(future).drop('Date',1).mean(0)
                futmean = futmean.to_dict()
                futmean['Rain'] = 0.0
                futmean.update({'MorP':'P'})
                futmean.update({'Date':mykey})
                wthdata.append(futmean)
            tcurrent = tcurrent + tdelta
        self.wdata = pd.DataFrame(wthdata)
        self.wdata.drop_duplicates(subset='Date',keep='first',
                                   inplace=True)
        self.wdata.set_index('Date',inplace=True)
        self.wdata.index.name = None
        indx = len(self.wdata.columns)-1
        self.wdata.insert(indx,'ETref',float('NaN'))

        #If needed and wanted, update wdata with 7-day NDFD forecast
        if needfuture and usefc:
            fc = Forecast(33.069,-111.972,wndht=self.wndht)
            fc.getforecast()
            for i in list(range(-1,10)):
                day = today + datetime.timedelta(days=i)
                key = day.strftime('%Y-%j')
                for item in ['Tmax','Tmin','Tdew','Wndsp']:
                    val = fc.forecast.loc[key,item]
                    if not math.isnan(val):
                        self.wdata.loc[key,item] = val

        #Compute ASCE Standardized Reference ET
        for index, row in self.wdata.iterrows():
            etref = self.compute_etref(index)
            self.wdata.loc[index,'ETref'] = etref

        #Check for NaN
        lessvapr = self.wdata.drop('Vapr', axis=1)
        nanrows = lessvapr.isna().any(1).to_numpy().nonzero()[0]
        if len(nanrows) > 0:
            print('There were problems with the following records:')
            for item in nanrows:
                print(self.wdata.iloc[item])
        else:
            print('The weather data was loaded with no issue.')



"""
########################################################################
The soil_profile.py module contains the SoilProfile class, which
provides I/O tools for defining soil properties stratified by soil
layer. Layered soil properties can be used (optionally) in the model
class to simulate the water balance, assuming the user has data to
define the soil properties by layer. Alternatively, users can define
soil properties using three parameters in the Parameters class (theta0,
thetaFC, and thetaWP), but this method assumes a single homogenous soil
layer.

The soil_profile.py module contains the following:
    SoilProfile - A class for managing soil properties stratified
       by soil layer

08/10/2022 Initial Python functions developed by Josh Brekel, USDA-ARS
09/27/2022 Finalized updates for inclusion in the pyfao56 Python package
########################################################################
"""

import pandas as pd
import datetime

class SoilProfile:
    """A class for managing layered soil profile data in pyfao56.

    Attributes
    ----------
    comment : str, optional
        User-defined file descriptions or metadata (default = '')
    tmstmp : datetime
        Time stamp for the class
    cnames : list
        Column names for sdata
    sdata : DataFrame
        Soil profile data as float
        index = Bottom depth of the layer as integer (cm)
        columns = ['thetaFC', 'thetaWP', 'theta0']
            thetaFC - Volumetric Soil Water Content, Field Capacity
                      (cm3/cm3)
            thetaWP - Volumetric Soil Water Content, Wilting Point
                      (cm3/cm3)
            theta0  - Volumetric Soil Water Content, Initial
                      (cm3/cm3)

    Methods
    -------
    savefile(filepath='pyfao56.sol')
        Save the soil profile data to a file
    loadfile(filepath='pyfao56.sol')
        Load soil profile data from a file
    customload()
        Users can override for custom loading of soil data.
    """

    def __init__(self, filepath=None, comment=''):
        """Initialize the SoilProfile class attributes.

        If filepath is provided, soil data is loaded from the file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = None).
        comment : str, optional
            User-defined file descriptions or metadata (default = '')
        """

        self.comment = 'Comments: ' + comment.strip()
        self.tmstmp = datetime.datetime.now()
        self.cnames = ['thetaFC', 'thetaWP', 'theta0']
        self.sdata = pd.DataFrame(columns=self.cnames)

        if filepath is not None:
            self.loadfile(filepath)

    def __str__(self):
        """Represent the SoilProfile class as a string"""

        self.tmstmp = datetime.datetime.now()
        timestamp = self.tmstmp.strftime('%m/%d/%Y %H:%M:%S')
        fmts = {'__index__':'{:5d}'.format,
                'thetaFC'  :'{:7.3f}'.format,
                'thetaWP'  :'{:7.3f}'.format,
                'theta0'   :'{:7.3f}'.format}
        ast ='*'*72
        s = ('{:s}\n'
             'pyfao56: FAO-56 Evapotranspiration in Python\n'
             'Soil Profile Data\n'
             'Timestamp: {:s}\n'
             '{:s}\n'
             '{:s}\n'
             '{:s}\n'
             'Depth').format(ast,timestamp,ast,self.comment,ast)
        for cname in self.cnames:
            s += '{:>8s}'.format(cname)
        s += '\n'
        if not self.sdata.empty:
            s += self.sdata.to_string(header=False,
                                      na_rep='    NaN',
                                      formatters=fmts)
        return s

    def savefile(self, filepath='pyfao56.sol'):
        """Save pyfao56 soil profile data to a file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.sol')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'w')
        except FileNotFoundError:
            print('The filepath for soil profile data is not found.')
        else:
            f.write(self.__str__())
            f.close()

    def loadfile(self, filepath='pyfao56.sol'):
        """Load pyfao56 soil profile data from a file.

        Parameters
        ----------
        filepath : str, optional
           Any valid filepath string (default = 'pyfao56.sol')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'r')
        except FileNotFoundError:
            print('The filepath for soil profile data is not found.')
        else:
            lines = f.readlines()
            f.close()
            ast = '*' * 72
            a = [i for i,line in enumerate(lines) if line.strip()==ast]
            endast = a[-1] 
            if endast == 3: #v1.1.0 and prior - no timestamps & metadata
                self.comment = 'Comments: '
            else:
                self.comment = ''.join(lines[5:endast]).strip()
            if endast >= 4:
                ts = lines[3].strip().split('stamp:')[1].strip()
                ts = datetime.datetime.strptime(ts,'%m/%d/%Y %H:%M:%S')
                self.tmstmp = ts
            for line in lines[endast+2:]:
                line = line.strip().split()
                depth = int(line[0])
                data = list()
                for i in list(range(1, 4)):
                    data.append(float(line[i]))
                self.sdata.loc[depth] = data

    def customload(self):
        """Override this function to customize loading soil data."""

        pass



"""
########################################################################
The weather.py module contains the Weather class, which provides I/O
tools for defining weather input data, as required for FAO-56
calculations.

The weather.py module contains the following:
    Weather - A class for managing weather data for FAO-56 calculations

01/07/2016 Initial Python functions developed by Kelly Thorp
11/04/2021 Finalized updates for inclusion in the pyfao56 Python package
08/03/2022 Added an input variable for measured vapor pressure
########################################################################
"""

import pandas as pd
from pyfao56 import refet
import datetime

class Weather:
    """A class for managing weather data for FAO-56 calculations.

    Attributes
    ----------
    comment : str, optional
        User-defined file descriptions or metadata (default = '')
    tmstmp : datetime
        Time stamp for the class
    rfcrp : str
        Type of reference crop  - Short ('S') or Tall ('T')
    z : float
        Weather station elevation (z) (m)
    lat : float
        Weather station latitude (decimal degrees)
    wndht : float
        Weather station wind speed measurement height (m)
    cnames : list
        Column names for wdata
    wdata : DataFrame
        Weather data as float
        index - Year and day of year as string ('yyyy-ddd')
        columns - ['Srad','Tmax','Tmin','Vapr','Tdew','RHmax','RHmin',
                   'Wndsp','Rain','ETref','MorP']
            Srad  - Incoming solar radiation (MJ/m2)
            Tmax  - Daily maximum air temperature (deg C)
            Tmin  - Daily minimum air temperature (deg C)
            Vapr  - Daily average vapor pressure (kPa)
            Tdew  - Daily average dew point temperature (deg C)
            RHmax - Daily maximum relative humidity (%)
            RHmin - Daily minimum relative humidity (%)
            Wndsp - Daily average wind speed (m/s)
            Rain  - Daily precipitation (mm)
            ETref - Daily reference ET (mm)
            MorP  - Measured ('M') or Predicted ('P') data

    Methods
    -------
    savefile(filepath='pyfao56.wth')
        Save the weather data to a file
    loadfile(filepath='pyfao56.wth')
        Load the weather data from a file
    customload()
        Users can override for custom weather loading, for example from
        meteorological network webpages
    compute_etref(index)
        Compute ASCE standardized reference ET for the weather data at
        index in self.wdata
    """

    def __init__(self,filepath=None,comment=''):
        """Initialize the Weather class attributes.

        If filepath is provided, weather data is loaded from the file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = None)
        comment : str, optional
            User-defined file descriptions or metadata (default = '')
        """

        self.comment = 'Comments: ' + comment.strip()
        self.tmstmp = datetime.datetime.now()
        self.rfcrp = 'S'
        self.z     = float('NaN')
        self.lat   = float('NaN')
        self.wndht = float('NaN')
        self.cnames = ['Srad','Tmax','Tmin','Vapr','Tdew','RHmax',
                       'RHmin','Wndsp','Rain','ETref','MorP']
        self.wdata = pd.DataFrame(columns=self.cnames)

        if filepath is not None:
            self.loadfile(filepath)

    def __str__(self):
        """Represent the Weather class variables as a string."""

        self.tmstmp = datetime.datetime.now()
        timestamp = self.tmstmp.strftime('%m/%d/%Y %H:%M:%S')
        fmts = {'Srad':'{:6.2f}'.format,'Tmax':'{:6.2f}'.format,
                'Tmin':'{:6.2f}'.format,'Tdew':'{:6.2f}'.format,
                'Vapr':'{:6.2f}'.format,'RHmax':'{:6.2f}'.format,
                'RHmin':'{:6.2f}'.format,'Wndsp':'{:6.2f}'.format,
                'Rain':'{:6.2f}'.format,'ETref':'{:6.2f}'.format,
                'MorP':'{:>5s}'.format}
        ast='*'*72
        s = ('{:s}\n'
             'pyfao56: FAO-56 Evapotranspiration in Python\n'
             'Weather Data\n'
             'Timestamp: {:s}\n'
             '{:s}\n'
             '{:s}\n'
             '{:s}\n'
             '{:>12s} Reference crop - Short (\'S\') or Tall (\'T\')\n'
             '{:12.7f} Weather station elevation (z) (m)\n'
             '{:12.7f} Weather station latitude (decimal degrees)\n'
             '{:12.7f} Wind speed measurement height (m)\n\n'
             'Daily weather data:\n'
             'Year-DOY'
             ).format(ast,timestamp,ast,self.comment,ast,self.rfcrp,
                      self.z,self.lat,self.wndht)
        for cname in self.cnames:
            s += '{:>7s}'.format(cname)
        s += '\n'
        if not self.wdata.empty:
            s += self.wdata.to_string(header=False,na_rep='   NaN',
                                      formatters=fmts)
        return s

    def savefile(self,filepath='pyfao56.wth'):
        """Save pyfao56 weather data to a file.

        Parameters
        ----------
        filepath : str, optional
           Any valid filepath string (default = 'pyfao56.wth')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'w')
        except FileNotFoundError:
            print('The filepath for weather data is not found.')
        else:
            f.write(self.__str__())
            f.close()

    def loadfile(self, filepath='pyfao56.wth'):
        """Load pyfao56 weather data from a file.

        Parameters
        ----------
        filepath : str, optional
           Any valid filepath string (default = 'pyfao56.wth')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'r')
        except FileNotFoundError:
            print('The filepath for weather data is not found.')
        else:
            lines = f.readlines()
            f.close()
            ast = '*' * 72
            a = [i for i,line in enumerate(lines) if line.strip()==ast]
            endast = a[-1] 
            if endast == 3: #v1.1.0 and prior - no timestamps & metadata
                self.comment = 'Comments: '
            else:
                self.comment = ''.join(lines[5:endast]).strip()
            if endast >= 4:
                ts = lines[3].strip().split('stamp:')[1].strip()
                ts = datetime.datetime.strptime(ts,'%m/%d/%Y %H:%M:%S')
                self.tmstmp = ts
            self.rfcrp = lines[endast+1][:12].strip()
            self.z = float(lines[endast+2][:12])
            self.lat = float(lines[endast+3][:12])
            self.wndht = float(lines[endast+4][:12])
            self.wdata = pd.DataFrame(columns=self.cnames)
            for line in lines[endast+8:]:
                line = line.strip().split()
                year = line[0][:4]
                doy = line[0][-3:]
                key = '{:04d}-{:03d}'.format(int(year),int(doy))
                data = list()
                for i in list(range(1,11)):
                    data.append(float(line[i]))
                data.append(line[11].strip())
                self.wdata.loc[key] = data

    def customload(self):
        """Override this function to customize loading weather data."""

        pass

    def compute_etref(self,index):
        """Compute ASCE standardized reference ET for data at index.

        Parameters
        ----------
        index : str
            The Year-DOY ('yyyy-ddd') index of self.wdata

        Returns
        -------
        ETref : float
            Daily standardized reference evapotranspiration for the
            short or tall reference crop (mm)
        """

        ETref = refet.ascedaily(self.rfcrp,
                                self.z,
                                self.lat,
                                float(index[-3:]),#DOY
                                self.wdata.loc[index,'Srad'],
                                self.wdata.loc[index,'Tmax'],
                                self.wdata.loc[index,'Tmin'],
                                self.wdata.loc[index,'Vapr'],
                                self.wdata.loc[index,'Tdew'],
                                self.wdata.loc[index,'RHmax'],
                                self.wdata.loc[index,'RHmin'],
                                self.wdata.loc[index,'Wndsp'],
                                self.wndht)
        return ETref



"""
########################################################################
The autoirrigate.py module contains the AutoIrrigate class, which
provides I/O tools for defining conditions for scheduling irrigation
automatically in pyfao56.

The autoirrigate.py module contains the following:
    AutoIrrigate - A class for managing multiple sets of conditions for
                   automatic irrigation

12/08/2023 Initial Python framework established for auto irrigation
02/08/2024 Further development, ideas from Kendall DeJonge & Fared Farag
02/14/2024 Finalized updates for inclusion in the pyfao56 Python package
########################################################################
"""

import pandas as pd
import datetime

class AutoIrrigate:
    """A class for managing multiple sets of autoirrigation conditions

    Attributes
    ----------
    comment : str, optional
        User-defined file descriptions or metadata (default = '')
    tmstmp : datetime
        Time stamp for the class
    aidata : DataFrame
        AutoIrrigate data with mixed data types
        index - counter as int
        columns - ['start','end','alre','idow','fpdep','fpday','fpact',
                   'mad','madDr','ksc','dsli','dsle','evnt','ifix',
                   'itdr','itfdr','ietrd','ietri','ietre','ettyp',
                   'iper','ieff','imin','imax','fw']
            Variables to determine if autoirrigation occurs or not:
            start - Autoirrigate only on start date or later
                    (str, 'yyyy-ddd')
            end   - Autoirrigate only on end date or earlier
                    (str, 'yyyy-ddd')
            alre  - Autoirrigate only after last reported irrigation
                    event (boolean)
            idow  - Autoirrigate only on specified days of the week
                    (str, '0123456')
                    (0:sun, 1:mon, 2:tue, 3:wed, 4:thu, 5:fri, 6:sat)
            fpdep - Threshold for forcasted precipitation depth
                    (float, mm)
            fpday : Number of days to consider forecasted precipitation
                    (float, days)
            fpact : Action if forecasted precip is above threshold
                    'proceed' - Proceed to autoirrigate anyway if needed
                    'cancel'  - Do not autoirrigate
                    'reduce'  - Deduct forecasted precip from
                                autoirrigation amount
            mad   - Management allowed depletion (float, mm/mm)
                    Autoirrigate if fractional root-zone soil water
                    depletion (fDr) >= mad
            madDr - Management allowed depletion (float, mm)
                    Autoirrigate if root-zone soil water depletion (Dr)
                    >= madDr
            ksc   - Critical value for transpiration reduction factor Ks
                    (float, 0-1, 1:full transpiration, 0:no trans.)
                    Autoirrigate if Ks <= ksc
            dsli  - Days since last irrigation event
                    (float, days)
                    Autoirrigate if days since last irrigation >= dsli
            dsle  - Days since last watering event, considering both
                    eff. precip and eff. irrigation (float, days)
                    Autoirrigate if days since last watering event
                    >= dsle
            evnt  - Minimum depth of effective precipitation and
                    effective irrigation to be considered a watering
                    event (float, mm)

            The default autoirrigation amount is root-zone soil water
            depletion (Dr, mm). Variables to alter this irrigation
            amount are as follows:

            ifix  - Apply a fixed autoirrigation amount (float, mm)
            itdr  - Target a specfic root-zone soil water depletion (Dr)
                    following autoirrigation (float, mm)
            itfdr - Target a specific fractional root-zone soil water
                    depletion following autoirrigation (float, mm/mm)
            ietrd - Replace ET minus effective precipitation from the
                    past given number of days (float, days)
            ietri - Replace ET minus effective precipitation since the
                    last irrigation event (boolean)
            ietre - Replace ET minus effective precipitation since the
                    last watering event (boolean)
            ettyp - Specify type of ET to replace
                    'ETa'  - Replace ETa less precip
                    'ETcm' - Replace ETcm less precip
                    'ETc'  - Replace ETc less precip
            iper  - Adjust the autoirrigation amount by a fixed
                    percentage (float, %)
            ieff  - Consider an application efficiency for
                    autoirrigation (float, %)
            imin  - Limit autoirrigation to >= minimum amount (float,mm)
            imax  - Limit autoirrigation to <= maximum amount (float,mm)

            fw    - Fraction of soil surface wetted (FAO-56 Table 20)

    Methods
    -------
    savefile(filepath='pyfao56.ati')
        Save the autoirrigate parameters to a file
    loadfile(filepath='pyfao56.ati')
        Load the autoirrigate parameters from a file
    addset(start,end,alre=True,idow='0123456',fpdep=25.,fpday=0,
           fpact='proceed',mad=NaN,madDr=NaN,ksc=NaN,dsli=NaN,dsle=NaN,
           evnt=10.,ifix=NaN,itdr=NaN,itfdr=NaN,ietrd=NaN,ietri=NaN,
           ietre=NaN,ettyp='ETa',iper=100.,ieff=100.,imin=NaN,imax=NaN,
           fw=1.)
        Add a set of autoirrigation parameters to self.aidata
    removeset(index)
        Remove a set of autoirrigation parameters from self.aidata
    customload()
        Override this function to customize loading of autoirrigation
        conditions
    """

    def __init__(self,filepath=None,comment=''):
        """Initialize the AutoIrrigate class attributes.

        If filepath is provided, data is loaded from the file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = None)
        comment : str, optional
            User-defined file descriptions or metadata (default = '')
        """

        self.comment = 'Comments: ' + comment.strip()
        self.tmstmp = datetime.datetime.now()
        self.cnames = ['start','end','alre','idow','fpdep','fpday',
                       'fpact','mad','madDr','ksc','dsli','dsle','evnt',
                       'ifix','itdr','itfdr','ietrd','ietri','ietre',
                       'ettyp','iper','ieff','imin','imax','fw']
        self.aidata = pd.DataFrame(columns=self.cnames)

        if filepath is not None:
            self.loadfile(filepath)

    def __str__(self):
        """Represent the AutoIrrigate class variables as a string."""

        self.tmstmp = datetime.datetime.now()
        timestamp = self.tmstmp.strftime('%m/%d/%Y %H:%M:%S')
        fmts = {'start':'{:>8s}'.format ,'end'  :'{:>8s}'.format ,
                'alre' :'{!s:>5}'.format,'idow' :'{:>7s}'.format ,
                'fpdep':'{:6.2f}'.format,'fpday':'{:6.0f}'.format,
                'fpact':'{:>7s}'.format ,'mad'  :'{:6.3f}'.format,
                'madDr':'{:6.2f}'.format,'ksc'  :'{:6.3f}'.format,
                'dsli' :'{:6.0f}'.format,'dsle' :'{:6.0f}'.format,
                'evnt' :'{:6.2f}'.format,'ifix' :'{:6.2f}'.format,
                'itdr' :'{:6.2f}'.format,'itfdr':'{:6.3f}'.format,
                'ietrd':'{:6.0f}'.format,'ietri':'{!s:>5}'.format,
                'ietre':'{!s:>5}'.format,'ettyp':'{!s:>6}'.format,
                'iper' :'{:6.2f}'.format,'ieff' :'{:6.2f}'.format,
                'imin' :'{:6.2f}'.format,'imax' :'{:6.2f}'.format,
                'fw'   :'{:6.2f}'.format}
        fmthead = ['  {:>8s}','  {:>8s}','  {:>5s}','  {:>7s}',
                    ' {:>6s}', ' {:>6s}','  {:>7s}', ' {:>6s}',
                    ' {:>6s}', ' {:>6s}', ' {:>6s}', ' {:>6s}',
                    ' {:>6s}', ' {:>6s}', ' {:>6s}', ' {:>6s}',
                    ' {:>6s}','  {:>5s}','  {:>5s}','  {:>6s}',
                    ' {:>6s}', ' {:>6s}', ' {:>6s}', ' {:>6s}',
                    ' {:>6s}']
        ast='*'*72
        s=('{:s}\n'
           'pyfao56: FAO-56 Evapotranspiration in Python\n'
           'AutoIrrigate Data\n'
           'Timestamp: {:s}\n'
           '{:s}\n'
           '{:s}\n'
           '{:s}\n'
          ).format(ast,timestamp,ast,self.comment,ast)
        s += ' '*len(str(self.aidata.shape[0]))
        for i, cname in enumerate(self.cnames):
            s += fmthead[i].format(cname)
        s += '\n'
        if not self.aidata.empty:
            s += self.aidata.to_string(header=False,na_rep='   NaN',
                                       formatters=fmts)
        return s

    def savefile(self,filepath='pyfao56.ati'):
        """Save pyfao56 autoirrigate parameters to a file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.ati')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'w')
        except FileNotFoundError:
            print('The filepath for autoirrigate data is not found.')
        else:
            f.write(self.__str__())
            f.close()

    def loadfile(self, filepath='pyfao56.ati'):
        """Load pyfao56 autoirrigate parameters from a file.

        Parameters
        ----------
        filepath : str, optional
            Any valid filepath string (default = 'pyfao56.ati')

        Raises
        ------
        FileNotFoundError
            If filepath is not found.
        """

        try:
            f = open(filepath, 'r')
        except FileNotFoundError:
            print('The filepath for autoirrigate data is not found.')
        else:
            lines = f.readlines()
            f.close()
            ast = '*' * 72
            a = [i for i,line in enumerate(lines) if line.strip()==ast]
            endast = a[-1]
            if endast == 3: #v1.1.0 and prior - no timestamps & metadata
                self.comment = 'Comments: '
            else:
                self.comment = ''.join(lines[5:endast]).strip()
            if endast >= 4:
                ts = lines[3].strip().split('stamp:')[1].strip()
                ts = datetime.datetime.strptime(ts,'%m/%d/%Y %H:%M:%S')
                self.tmstmp = ts
            self.aidata = pd.DataFrame(columns=self.cnames)
            for i, line in enumerate(lines[endast+2:]):
                line = line.strip().split()
                data = list()
                data.append(line[1])          #start
                data.append(line[2])          #end
                data.append(line[3]=='True')  #alre
                data.append(line[4])          #idow
                data.append(float(line[5]))   #fpdep
                data.append(float(line[6]))   #fpday
                data.append(line[7])          #fpact
                data.append(float(line[8]))   #mad
                data.append(float(line[9]))   #madDr
                data.append(float(line[10]))  #ksc
                data.append(float(line[11]))  #dsli
                data.append(float(line[12]))  #dsle
                data.append(float(line[13]))  #evnt
                data.append(float(line[14]))  #ifix
                data.append(float(line[15]))  #itdr
                data.append(float(line[16]))  #itfdr
                data.append(float(line[17]))  #ietrd
                data.append(line[18]=='True') #ietri
                data.append(line[19]=='True') #ietre
                data.append(line[20].strip()) #ettyp
                data.append(float(line[21]))  #iper
                data.append(float(line[22]))  #ieff
                data.append(float(line[23]))  #imax
                data.append(float(line[24]))  #imin
                data.append(float(line[25]))  #fw
                self.aidata.loc[i] = data

    def addset(self,start,end,alre=True,idow='0123456',fpdep=25.,
               fpday=0,fpact='proceed',mad=float('NaN'),
               madDr=float('NaN'),ksc=float('NaN'),dsli=float('NaN'),
               dsle=float('NaN'),evnt=10.,ifix=float('NaN'),
               itdr=float('NaN'),itfdr=float('NaN'),ietrd=float('NaN'),
               ietri=False,ietre=False,ettyp='ETa',iper=100.,ieff=100.,
               imin=0.,imax=float('NaN'),fw=1.):
        """Add a set of autoirrigation parameters to aidata

        Default parameter values are given below. Users should update
        the parameters with values for their desired autoirrigation
        conditions.

        Parameters
        ----------
        See AutoIrrigate class docstring for parameter definitions.
        start : str, start year and doy for parameter set ('yyyy-ddd')
        end   : str, end year and doy for parameter set ('yyy-ddd')
        alre  : boolean, optional, default=True
        idow  : str    , optional, default='0123456'
        fpdep : float  , optional, default=25.
        fpday : float  , optional, default=0
        fpact : str    , optional, default='proceed'
        mad   : float  , optional, default=NaN
        madDr : float  , optional, default=NaN
        ksc   : float  , optional, default=NaN
        dsli  : float  , optional, default=NaN
        dsle  : float  , optional, default=NaN
        evnt  : float  , optional, default=10.
        ifix  : float  , optional, default=NaN
        itdr  : float  , optional, default=NaN
        itfdr : float  , optional, default=NaN
        ietrd : float  , optional, default=NaN
        ietri : boolean, optional, default=False
        ietre : boolean, optional, default=False
        ettyp : string , optional, default='ETa'
        iper  : float  , optional, default=100.
        ieff  : float  , optional, default=100.
        imin  : float  , optional, default=0.
        imax  : float  , optional, default=NaN
        fw    : float  , optional, default=1.
        """

        i = len(self.aidata)
        data = [str(start),str(end),bool(alre),str(idow),float(fpdep),
                float(fpday),str(fpact),float(mad),float(madDr),
                float(ksc),float(dsli),float(dsle),float(evnt),
                float(ifix),float(itdr),float(itfdr),float(ietrd),
                bool(ietri),bool(ietre),str(ettyp),float(iper),
                float(ieff),float(imin),float(imax),float(fw)]
        self.aidata.loc[i] = data

    def removeset(self,index):
        """Remove a set of autoirrigation parameters from aidata

        Parameters
        ----------
        index : int
            Index of the set of autoirrigation data to be removed
        """

        self.aidata.drop(index)

    def customload(self):
        """Override function to customize loading autoirrigate data."""

        pass
